"""
This test file tests the lib.tokenclass

The lib.tokenclass depends on the DB model and lib.user
"""
import logging

from testfixtures import LogCapture
from .base import MyTestCase, FakeAudit, FakeFlaskG
from privacyidea.lib.resolver import (save_resolver)
from privacyidea.lib.realm import (set_realm)
from privacyidea.lib.user import (User)
from privacyidea.lib.tokens.totptoken import TotpTokenClass
from privacyidea.lib.policy import (PolicyClass, set_policy, delete_policy, SCOPE)
from unittest import mock
from privacyidea.models import (Token,
                                 Config,
                                 Challenge)
from privacyidea.lib.config import (set_privacyidea_config, set_prepend_pin)
import datetime
import binascii
import time

PWFILE = "tests/testdata/passwords"


class TOTPTokenTestCase(MyTestCase):
    """
    Test the token on the database level
    """
    resolvername1 = "resolver1"
    resolvername2 = "Resolver2"
    resolvername3 = "reso3"
    realm1 = "realm1"
    realm2 = "realm2"
    serial1 = "SE123456"
    otpkey = "3132333435363738393031323334353637383930"

    """
    47251644    942826
    47251645    063321
    47251646    306773
    47251647    722053
    47251648    032819
    47251649    705493
    47251650    589836
    """

    # add_user, get_user, reset, set_user_identifiers

    def test_00_create_user_realm(self):
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1, [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        user = User(login="root",
                    realm=self.realm1,
                    resolver=self.resolvername1)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.resolver1@realm1>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm1', resolver='resolver1')"
        self.assertTrue(user_repr == expected, user_repr)

    def test_01_create_token(self):
        db_token = Token(self.serial1, tokentype="totp")
        db_token.save()
        token = TotpTokenClass(db_token)
        self.assertTrue(token.token.serial == self.serial1, token)
        self.assertTrue(token.token.tokentype == "totp", token.token.tokentype)
        self.assertTrue(token.type == "totp", token)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "TOTP", class_prefix)
        self.assertTrue(token.get_class_type() == "totp", token)

    def test_02_set_user(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        self.assertTrue(token.token.tokentype == "totp",
                        token.token.tokentype)
        self.assertTrue(token.type == "totp", token.type)

        token.add_user(User(login="cornelius",
                            realm=self.realm1))
        self.assertEqual(token.token.first_owner.resolver, self.resolvername1)
        self.assertEqual(token.token.first_owner.user_id, "1000")

        user_object = token.user
        self.assertTrue(user_object.login == "cornelius",
                        user_object)
        self.assertTrue(user_object.resolver == self.resolvername1,
                        user_object)

    def test_03_reset_failcounter(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.token.failcount = 10
        token.reset()
        self.assertTrue(token.token.failcount == 0,
                        token.token.failcount)

    def test_04_base_methods(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        self.assertTrue(token.check_otp("123456", 1, 10) == -1)

        c = token.create_challenge("transactionid")
        self.assertTrue(c[0], c)
        self.assertTrue("transactionid" in c[2], c)

        # set the description
        token.set_description("something new")
        self.assertTrue(token.token.description == "something new",
                        token.token)

        # set defaults
        token.set_defaults()
        self.assertTrue(token.token.otplen == 6)
        self.assertTrue(token.token.sync_window == 1000)

        token.resync("1234", "3456")

        token.token.count_window = 17
        self.assertTrue(token.get_otp_count_window() == 17)

        token.token.count = 18
        self.assertTrue(token.get_otp_count() == 18)

        token.token.active = False
        self.assertTrue(token.is_active() is False)

        token.token.failcount = 7
        self.assertTrue(token.get_failcount() == 7)
        token.set_failcount(8)
        self.assertTrue(token.token.failcount == 8)

        token.token.maxfail = 12
        self.assertTrue(token.get_max_failcount() == 12)

        self.assertEqual(token.get_user_id(), token.token.first_owner.user_id)

        self.assertTrue(token.get_serial() == "SE123456", token.token.serial)
        self.assertTrue(token.get_tokentype() == "totp",
                        token.token.tokentype)

        token.set_so_pin("sopin")
        token.set_user_pin("userpin")
        token.set_otpkey(self.otpkey)
        token.set_otplen(8)
        token.set_otp_count(1000)
        self.assertTrue(len(token.token.so_pin) == 32,
                        token.token.so_pin)
        self.assertTrue(len(token.token.user_pin) == 32,
                        token.token.user_pin)
        self.assertEqual(len(token.token.key_enc), 96, token.token.key_enc)
        self.assertTrue(token.get_otplen() == 8)
        self.assertTrue(token.token.count == 1000,
                        token.token.count)

        token.set_maxfail(1000)
        self.assertTrue(token.token.maxfail == 1000)

        token.set_count_window(52)
        self.assertTrue(token.get_count_window() == 52)

        token.set_sync_window(53)
        self.assertTrue(token.get_sync_window() == 53)

    def test_06_set_pin(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.set_pin("hallo")
        (ph1, pseed) = token.get_pin_hash_seed()
        # check the database
        token.set_pin("blubber")
        ph2 = token.token.pin_hash
        self.assertTrue(ph1 != ph2)
        token.set_pin_hash_seed(ph1, pseed)

    def test_07_enable(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.enable(False)
        self.assertTrue(token.token.active is False)
        token.enable()
        self.assertTrue(token.token.active)

    def test_05_get_set_realms(self):
        set_realm(self.realm2)
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        realms = token.get_realms()
        self.assertTrue(len(realms) == 1, realms)
        token.set_realms([self.realm1, self.realm2])
        realms = token.get_realms()
        self.assertTrue(len(realms) == 2, realms)

    def test_99_delete_token(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.delete_token()

        db_token = Token.query.filter_by(serial=self.serial1).first()
        self.assertTrue(db_token is None, db_token)

    def test_08_info(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.set_hashlib("sha1")
        ti = token.get_tokeninfo()
        self.assertTrue("hashlib" in ti, ti)

    def test_09_failcount(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        start = token.token.failcount
        end = token.inc_failcount()
        self.assertTrue(end == start + 1, (end, start))

    def test_10_get_hashlib(self):
        # check if functions are returned
        for hl in ["sha1", "md5", "sha256", "sha512",
                   "sha224", "sha384", "", None]:
            self.assertTrue(hasattr(TotpTokenClass.get_hashlib(hl),
                                    '__call__'),
                            TotpTokenClass.get_hashlib(hl))

    def test_11_tokeninfo(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.add_tokeninfo("key1", "value2")
        info1 = token.get_tokeninfo()
        self.assertTrue("key1" in info1, info1)
        token.add_tokeninfo("key2", "value3")
        info2 = token.get_tokeninfo()
        self.assertTrue("key2" in info2, info2)
        token.set_tokeninfo(info1)
        info2 = token.get_tokeninfo()
        self.assertTrue("key2" not in info2, info2)
        self.assertTrue(token.get_tokeninfo("key1") == "value2",
                        info2)

        # auth counter
        token.set_count_auth_success_max(200)
        token.set_count_auth_max(1000)
        token.set_count_auth_success(100)
        token.inc_count_auth_success()
        token.set_count_auth(200)
        token.inc_count_auth()
        self.assertTrue(token.get_count_auth_success_max() == 200)
        self.assertTrue(token.get_count_auth_success() == 101)
        self.assertTrue(token.get_count_auth_max() == 1000)
        self.assertTrue(token.get_count_auth() == 201)

        self.assertTrue(token.check_auth_counter())
        token.set_count_auth_max(10)
        self.assertFalse(token.check_auth_counter())
        token.set_count_auth_max(1000)
        token.set_count_auth_success_max(10)
        self.assertFalse(token.check_auth_counter())

    def test_12_inc_otp_counter(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)

        token.set_otp_count(10)
        self.assertTrue(token.token.count == 10, token.token.count)
        # increase counter by 1
        token.inc_otp_counter()
        self.assertTrue(token.token.count == 11, token.token.count)
        # increase counter to 21
        Config(Key="DefaultResetFailCount", Value=True).save()
        token.inc_otp_counter(counter=20)
        self.assertTrue(token.token.count == 21, token.token.count)

    @mock.patch('time.time', mock.MagicMock(return_value=1686902767))
    def test_98_get_otp(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.update({"otpkey": self.otpkey,
                      "pin": "test",
                      "otplen": 6,
                      "timeShift": 0})
        counter = token._time2counter(time.time(), timeStepping=30)
        otp_now = token._calc_otp(counter)
        otp = token.get_otp()
        res = token.check_otp_exist(otp[2])
        self.assertEqual(res, counter)
        self.assertEqual(otp[2], otp_now)

    def test_13_check_otp(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.update({"otpkey": self.otpkey,
                      "pin": "test",
                      "otplen": 6})
        token.set_otp_count(47251644)
        # OTP does not exist
        self.assertTrue(token.check_otp("222333") == -1)
        # OTP does exist
        res = token.check_otp("722053", options={"initTime": 47251645 * 30})
        # Found the counter 47251647
        self.assertEqual(47251647, res)
        with LogCapture(level=logging.INFO) as lc:
            self.assertEqual(-1, token.check_otp("1234567"))
            lc.check_present(
                ('privacyidea.lib.decorators', 'INFO',
                 f'OTP value for token {token.token.serial} (type: {token.type}) '
                 f'has wrong length (7 != 6)'))
            self.assertEqual(-1, token.check_otp("12345"))
            lc.check_present(
                ('privacyidea.lib.decorators', 'INFO',
                 f'OTP value for token {token.token.serial} (type: {token.type}) '
                 f'has wrong length (5 != 6)'))

    def test_14_split_pin_pass(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)

        token.token.otplen = 6
        # postpend pin
        set_prepend_pin(False)
        _res, pin, value = token.split_pin_pass("222333test")
        self.assertTrue(pin == "test", pin)
        self.assertTrue(value == "222333", value)
        # prepend pin
        set_prepend_pin()
        _res, pin, value = token.split_pin_pass("test222333")
        self.assertTrue(pin == "test", pin)
        self.assertTrue(value == "222333", value)

    def test_15_check_pin(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.set_pin("test")
        self.assertTrue(token.check_pin("test"))
        self.assertFalse(token.check_pin("wrong pin"))

    def test_16_init_detail(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.add_init_details("otpkey", "11223344556677889900")
        detail = token.get_init_detail()
        self.assertTrue("otpkey" in detail, detail)
        # but the otpkey must not be written to token.token.info (DB)
        # As this only writes the OTPkey to the internal init_details dict
        self.assertTrue("otpkey" not in token.token.get_info(),
                        token.token.get_info())

        # Now get the Google Authenticator URL, which we only
        # get, if a user is specified.
        detail = token.get_init_detail(user=User("cornelius",
                                                 self.realm1))
        self.assertTrue("otpkey" in detail, detail)
        otpkey = detail.get("otpkey")
        self.assertTrue("img" in otpkey, otpkey)
        self.assertTrue("googleurl" in detail, detail)
        # some other stuff.
        self.assertRaises(Exception, token.set_init_details, "invalid value")
        token.set_init_details({"detail1": "value1"})
        self.assertTrue("detail1" in token.get_init_details(),
                        token.get_init_details())

    def test_17_update_token(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        # Failed update: genkey wrong
        self.assertRaises(Exception,
                          token.update,
                          {"description": "new desc",
                           "genkey": "17"})
        # genkey and otpkey used at the same time
        token.update({"otpkey": self.otpkey,
                      "genkey": "1"})

        token.update({"otpkey": self.otpkey,
                      "pin": "654321",
                      "otplen": 6})
        self.assertTrue(token.check_pin("654321"))
        self.assertTrue(token.token.otplen == 6)
        # update hashlib
        token.update({"otpkey": self.otpkey,
                      "hashlib": "sha1"})
        self.assertTrue(token.get_tokeninfo("hashlib") == "sha1",
                        token.get_tokeninfo())

        # save pin encrypted
        token.update({"genkey": 1,
                      "pin": "secret",
                      "encryptpin": "true"})
        # check if the PIN is encrypted
        self.assertTrue(token.token.pin_hash.startswith("@@"),
                        token.token.pin_hash)

        # update token without otpkey raises an error
        self.assertRaises(Exception, token.update, {"description": "test"})

        # update time settings
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.update({"otpkey": self.otpkey,
                      "otplen": 6,
                      "timeShift": 10,
                      "timeWindow": 180,
                      "timeStep": 30
                      })

    def test_18_challenges(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        resp = token.is_challenge_response(user=User(login="cornelius",
                                                     realm=self.realm1),
                                           passw="test123456")
        self.assertFalse(resp, resp)

        transaction_id = "123456789"
        chall = Challenge(self.serial1, transaction_id=transaction_id, challenge="Who are you?")
        chall.save()
        resp = token.is_challenge_response(user=User(login="cornelius",
                                                     realm=self.realm1),
                                           passw="test123456",
                                           options={"transaction_id": transaction_id})
        self.assertTrue(resp, resp)

        # test if challenge is valid
        chall.is_valid()

    def test_19_pin_otp_functions(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        db_token.set_pin("test")
        token = TotpTokenClass(db_token)
        # check OTP according to RFC 4226
        token.update({"otpkey": self.otpkey})
        self.assertTrue(db_token.otplen == 6, 6)
        set_prepend_pin()
        res, pin, otp = token.split_pin_pass("test123456")
        self.assertTrue(pin == "test", pin)
        self.assertTrue(otp == "123456", otp)
        self.assertTrue(token.check_pin(pin), pin)
        # get the OTP value for counter 1417549521
        res = token.get_otp(time_seconds=1417549521)
        self.assertTrue(res[0] == 1, res)
        self.assertTrue(res[2] == "227747", res)

        check = token.check_otp("722053", counter=47251647)
        # The OTP 722053 is of counter 47251647
        self.assertTrue(check == 47251647, check)
        # The tokenclass saves the counter to the database
        self.assertTrue(token.token.count == 47251647, token.token.count)

        check = token.check_otp("705493", counter=47251648)
        # The OTP 705493 is of counter 47251649, but it matches also.
        self.assertTrue(check == 47251649, check)

        # successful authentication
        res = token.authenticate("test589836")
        # This is the OTP value of the counter=47251650
        self.assertTrue(res == (True, 47251650, None), res)

        # try the same OTP value again will fail!
        res = token.authenticate("test589836")
        # This is the OTP value of the counter=47251650
        self.assertTrue(res == (True, -1, None), res)

        res = token.get_multi_otp()
        self.assertTrue(res[0] is False, res)
        token.update({"otpkey": self.otpkey,
                      "otplen": 6,
                      "timeShift": 0})
        token.token.count = 0
        res = token.get_multi_otp(count=5)
        self.assertTrue(res[0], res)
        self.assertTrue(res[1] == "OK", res)
        self.assertTrue(len(res[2].get("otp")) == 5, res[2].get("otp"))

        # Simulate the server time
        res = token.get_multi_otp(count=5, timestamp=47251644 * 30)
        self.assertTrue(res[0], res)
        self.assertTrue(res[1] == "OK", res)
        self.assertTrue(len(res[2].get("otp")) == 5, res[2].get("otp"))
        self.assertTrue(47251648 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47251647 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47251646 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47251645 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47251644 in res[2].get("otp"), res[2].get("otp"))

        # Simulate the server time
        res = token.get_multi_otp(count=5, curTime=datetime.datetime(2014, 12, 12))
        self.assertTrue(res[0], res)
        self.assertTrue(res[1] == "OK", res)
        self.assertTrue(len(res[2].get("otp")) == 5, res[2].get("otp"))
        self.assertTrue(47278080 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47278081 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47278082 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47278083 in res[2].get("otp"), res[2].get("otp"))
        self.assertTrue(47278084 in res[2].get("otp"), res[2].get("otp"))

        # do some failing otp checks
        token.token.otplen = 5
        self.assertTrue(token.check_otp("705493", counter=47251648) == -1, res)
        token.token.otplen = 6

        # Previous OTP value used again
        token.token.otplen = 6
        # The OTP for this counter was already presented to the server
        token.token.count = 47251648
        # 47251647 -> 722053
        res = token.check_otp("722053", options={"initTime": 47251649 * 30})
        self.assertTrue(res == -1, res)

        # simple OTPs of current time
        ret, _, dct = token.get_multi_otp(1)
        self.assertTrue(ret)
        self.assertGreater(list(dct["otp"].keys())[0], 47251648)
        ret, _, dct = token.get_multi_otp(1, curTime=datetime.datetime.now())
        self.assertTrue(ret)
        self.assertGreater(list(dct["otp"].keys())[0], 47251648)

    def test_20_check_challenge_response(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        db_token.set_pin("test")
        token = TotpTokenClass(db_token)
        r = token.check_challenge_response(user=None,
                                           passw="123454")
        # check empty challenges
        self.assertTrue(r == -1, r)

        # create a challenge and match the transaction_id
        c = Challenge(self.serial1, transaction_id="mytransaction",
                      challenge="Blah, what now?")
        # save challenge to the database
        c.save()
        r = token.check_challenge_response(user=None,
                                           passw="123454",
                                           options={"state": "mytransaction"})
        # The challenge matches, but the OTP does not match!
        self.assertTrue(r == -1, r)

        # test the challenge janitor
        c1 = Challenge(self.serial1, transaction_id="t1", validitytime=0)
        c1.save()
        c2 = Challenge(self.serial1, transaction_id="t2", validitytime=0)
        c2.save()
        c3 = Challenge(self.serial1, transaction_id="t3", validitytime=100)
        c3.save()
        c4 = Challenge(self.serial1, transaction_id="t4", validitytime=100)
        c4.save()
        num = Challenge.query.filter(Challenge.serial == self.serial1).count()
        self.assertTrue(num >= 5, num)
        # We pass the third challenge as the valid challenge.
        # So 3 challenges will be deleted.
        token.challenge_janitor()
        # Now see if those challenges are deleted
        num1 = Challenge.query.filter(Challenge.transaction_id == "t1").count()
        num2 = Challenge.query.filter(Challenge.transaction_id == "t2").count()
        num3 = Challenge.query.filter(Challenge.transaction_id == "t3").count()
        num4 = Challenge.query.filter(Challenge.transaction_id == "t4").count()
        self.assertTrue(num1 == 0)
        self.assertTrue(num2 == 0)
        self.assertTrue(num3 == 1)
        self.assertTrue(num4 == 1)

    def test_21_get_class_info(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        ti = token.get_class_info()
        self.assertTrue(ti.get("type") == "totp", ti)
        self.assertTrue("policy" in ti, ti)
        self.assertTrue("title" in ti, ti)
        self.assertTrue("user" in ti, ti)

    def test_22_autosync(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        set_privacyidea_config("AutoResync", True)
        token.update({"otpkey": self.otpkey,
                      "otplen": 6})
        token.token.count = 47251640
        token.set_sync_window(10)
        # counter = 47251649 => otp = 705493, is out of sync
        r = token.check_otp("705493", window=30,
                            options={"initTime": 47251644 * 30})
        self.assertTrue(r == -1, r)
        # counter = 47251650 => otp = 389836, will be autosynced.
        r = token.check_otp("589836", window=30,
                            options={"initTime": 47251645 * 30})
        self.assertTrue(r == 47251650, r)

        # counter = 47251640 => otp = 166325 is an old OTP value
        # counter = 47251641 => otp = 432730 is an old OTP value
        # Autoresync with two times the same old OTP value must not work out!
        r = token.check_otp("166325", window=30,
                            options={"initTime": 47251644 * 30})
        self.assertTrue(r == -1, r)
        r = token.check_otp("166325", window=30,
                            options={"initTime": 47251645 * 30})
        self.assertTrue(r == -1, r)
        # Autoresync with two consecutive old OTP values must not work out!
        r = token.check_otp("166325", window=30,
                            options={"initTime": 47251644 * 30})
        self.assertTrue(r == -1, r)
        r = token.check_otp("432730", window=30,
                            options={"initTime": 47251645 * 30})
        self.assertTrue(r == -1, r)

        # Autosync with a gap in the next otp value will fail
        token.token.count = 47251640
        # Just try some bullshit config value
        set_privacyidea_config("AutoResyncTimeout", "totally not a number")
        # counter = 47251648 => otp = 032819, is out of sync
        r = token.check_otp("032819", window=30,
                            options={"initTime": 47251645 * 30})
        self.assertTrue(r == -1, r)
        # counter = 47251650 => otp = 589836, will NOT _autosync
        r = token.check_otp("589836", window=30,
                            options={"initTime": 47251645 * 30})
        self.assertTrue(r == -1, r)

        # TOTP has no dueDate / AutoResyncTimeout

        # No _autosync
        set_privacyidea_config("AutoResync", False)
        token.token.count = 47251640
        token.set_sync_window(10)
        # counter = 47251649 => otp = 705493, is out of sync
        r = token.check_otp("705493", window=30,
                            options={"initTime": 47251644 * 30})
        self.assertTrue(r == -1, r)
        # counter = 47251650 => otp = 389836, will not get autosynced.
        r = token.check_otp("589836", window=30,
                            options={"initTime": 47251645 * 30})
        self.assertTrue(r == -1, r)

    def test_23_resync(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.update({"otpkey": self.otpkey,
                      "otplen": 6})
        token.set_sync_window(1000)
        # 705493 -> 47251649
        # 589836 -> 47251650
        # So the token might be at time 47251650,
        # but the server time is 47251600

        # The server time is 2000*30 seconds further, the resync will fail
        # 47253650 - 47251650 = 2000 ticks away
        token.token.count = 47251400
        r = token.resync("705493", "589836",
                         options={"initTime": 47253650 * 30})
        self.assertFalse(r)
        # Successful resync
        token.token.count = 47251400
        # The server time is 200*30 seconds further
        # 47251850 - 47251650 = 200 ticks away
        r = token.resync("705493", "589836",
                         options={"initTime": 47251850 * 30})
        self.assertTrue(r is True, r)
        # resync fails
        token.token.count = 0
        self.assertFalse(token.resync("399871", "123456"))
        # resync fails, the two correct OTP values are outside of the sync
        # window
        token.token.count = 0
        token.set_sync_window(5)
        self.assertFalse(token.resync("399871", "520489"))

    def test_24_challenges(self):
        db_token = Token.query.filter_by(serial=self.serial1).first()
        token = TotpTokenClass(db_token)
        token.update({"otpkey": self.otpkey,
                      "otplen": 6})
        token.set_pin("test")
        token.token.count = 0
        token.set_sync_window(10)
        token.set_count_window(5)
        self.assertTrue(token.is_challenge_request("test"))

    def test_25_sha256_token(self):
        # taken from https://tools.ietf.org/html/rfc6238#appendix-B
        # sha256 with a 20 byte seed
        serial = "sha25T"
        db_token = Token(serial, tokentype="totp")
        db_token.save()
        token = TotpTokenClass(db_token)
        token.set_otpkey(binascii.hexlify(b"12345678901234567890"))
        token.set_hashlib("sha256")
        token.set_otplen(8)
        token.save()
        # get it from the database again
        # |  1111111111 |  2005-03-18  | 00000000023523ED | 67062674 | SHA256 |
        db_token = Token.query.filter_by(serial=serial).first()
        token = TotpTokenClass(db_token)
        r = token.check_otp("67062674", options={"initTime": 1111111111})
        self.assertTrue(r)

        # sha256 with a 32 byte seed
        serial = "sha256"
        db_token = Token(serial, tokentype="totp")
        db_token.save()
        token = TotpTokenClass(db_token)
        token.set_otpkey(binascii.hexlify(b"12345678901234567890123456789012"))
        token.set_hashlib("sha256")
        token.set_otplen(8)
        token.save()
        db_token = Token.query.filter_by(serial=serial).first()
        token = TotpTokenClass(db_token)
        r = token.check_otp("67062674", options={"initTime": 1111111111})
        self.assertTrue(r)

        # sha512 with a 20 byte seed
        serial = "sha512"
        db_token = Token(serial, tokentype="totp")
        db_token.save()
        token = TotpTokenClass(db_token)
        token.set_otpkey(binascii.hexlify(b"12345678901234567890"))
        token.set_hashlib("sha512")
        token.set_otplen(8)
        token.save()
        db_token = Token.query.filter_by(serial=serial).first()
        token = TotpTokenClass(db_token)
        r = token.check_otp("99943326", options={"initTime": 1111111111})
        self.assertTrue(r)

        # sha512 with a 64byte seed
        serial = "sha512b"
        db_token = Token(serial, tokentype="totp")
        db_token.save()
        token = TotpTokenClass(db_token)
        token.set_otpkey(binascii.hexlify(
            b"1234567890123456789012345678901234567890123456789012345678901234"))
        token.set_hashlib("sha512")
        token.set_otplen(8)
        token.save()
        db_token = Token.query.filter_by(serial=serial).first()
        token = TotpTokenClass(db_token)
        r = token.check_otp("93441116", options={"initTime": 1234567890})
        self.assertTrue(r)

    def test_26_get_setting_type(self):
        r = TotpTokenClass.get_setting_type("totp.hashlib")
        self.assertEqual(r, "public")
        r = TotpTokenClass.get_setting_type("totp.blabla")
        self.assertEqual(r, "")

    def test_26_is_previous_otp(self):
        # check if the OTP was used previously
        serial = "previous"
        db_token = Token(serial, tokentype="totp")
        db_token.save()
        token = TotpTokenClass(db_token)
        token.set_otpkey(self.otpkey)
        token.set_hashlib("sha1")
        token.set_otplen(6)
        # Authenticate with the current OTP value
        counter = token._time2counter(time.time(), timeStepping=30)
        otp_now = token._calc_otp(counter)
        r = token.check_otp(otp_now, window=180)
        self.assertEqual(r, counter)
        # Now we try several is_previous_otp and the timeShift must stay the same!
        ts0 = float(token.get_tokeninfo("timeShift"))
        self.assertTrue(-181 < ts0 < 181)
        # Too old
        r = token.is_previous_otp(token._calc_otp(counter - 3))
        self.assertEqual(r, False)
        ts = float(token.get_tokeninfo("timeShift"))
        self.assertEqual(ts, ts0)
        # The same OTP value
        r = token.is_previous_otp(otp_now)
        self.assertEqual(r, True)
        ts = float(token.get_tokeninfo("timeShift"))
        self.assertEqual(ts, ts0)
        # Future value
        r = token.is_previous_otp(token._calc_otp(counter + 8))
        self.assertEqual(r, False)
        ts = float(token.get_tokeninfo("timeShift"))
        self.assertEqual(ts, ts0)

    def test_27_get_default_settings(self):
        params = {}
        g = FakeFlaskG()
        g.audit_object = FakeAudit()
        g.logged_in_user = {"user": "hans",
                            "realm": "default",
                            "role": "user"}
        set_policy("pol1", scope=SCOPE.USER, action="totp_hashlib=sha256,"
                                                    "totp_timestep=60,"
                                                    "totp_otplen=8")
        g.policy_object = PolicyClass()
        p = TotpTokenClass.get_default_settings(g, params)
        self.assertEqual(p.get("otplen"), "8")
        self.assertEqual(p.get("hashlib"), "sha256")
        self.assertEqual(p.get("timeStep"), "60")
        delete_policy("pol1")

        # the same should work for admins
        g.logged_in_user = {"user": "admin",
                            "realm": "super",
                            "role": "admin"}
        set_policy("pol1", scope=SCOPE.ADMIN, action="totp_hashlib=sha512,"
                                                     "totp_timestep=60,"
                                                     "totp_otplen=8")
        g.policy_object = PolicyClass()
        p = TotpTokenClass.get_default_settings(g, params)
        self.assertEqual(p.get("otplen"), "8")
        self.assertEqual(p.get("hashlib"), "sha512")
        self.assertEqual(p.get("timeStep"), "60")
        # test check if there is no logged in user
        g.logged_in_user = None
        p = TotpTokenClass.get_default_settings(g, params)
        self.assertEqual(p, {})
        delete_policy("pol1")



"""v3.10.2: Drop unique constraint on tokencontainerrealm

Revision ID: eac770c0bbed
Revises: 69e7817b9863
Create Date: 2024-10-30 11:21:28.721316

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError

# revision identifiers, used by Alembic.
revision = 'eac770c0bbed'
down_revision = '69e7817b9863'


def upgrade():
    migration_context = op.get_context()

    try:
        if migration_context.dialect.name in ['sqlite']:
            # SQLite allows constraint without a name, so we have to use a naming convention to be able to drop it
            naming_convention = {
                "ix": "ix_%(column_0_label)s",
                "uq": "uq_%(table_name)s_%(column_0_name)s_%(column_1_name)s",
                "ck": "ck_%(table_name)s_`%(constraint_name)s`",
                "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
                "pk": "pk_%(table_name)s"
            }
            with op.batch_alter_table(
                    "tokencontainerrealm", naming_convention=naming_convention) as batch_op:
                batch_op.drop_constraint("uq_tokencontainerrealm_container_id_realm_id", type_="unique")

        else:
            # All other databases autogenerate a name for constraints, we just need to get this name to drop it
            bind = op.get_bind()
            insp = sa.inspect(bind.engine)
            uqs = insp.get_unique_constraints("tokencontainerrealm")
            columns = ["container_id", "realm_id"]
            for uq in uqs:
                if len(list(set(columns).intersection(uq['column_names']))) == 2:
                    op.drop_constraint(uq['name'], 'tokencontainerrealm', type_='unique')

    except (OperationalError, ProgrammingError) as exx:
        if "no such constraint" in str(exx.orig).lower():
            print("Unique constraint on tokencontainerrealm already removed.")
        else:
            print("Could not drop unique constraint on tokencontainerrealm.")
            print(exx)
    except ValueError as exx:
        if "no such constraint" in str(exx).lower():
            print("Unique constraint on tokencontainerrealm already removed.")
        else:
            print("Could not drop unique constraint on tokencontainerrealm.")
            print(exx)


def downgrade():
    pass



"""
This test file tests the lib.token methods.

The lib.token depends on the DB model and lib.user and
all lib.tokenclasses

This tests the token functions on an interface level

We start with simple database functions:

getTokens4UserOrSerial
gettokensoftype
getToken....
"""
from .base import MyTestCase
from privacyidea.lib.error import privacyIDEAError, ResourceNotFoundError

from privacyidea.lib.tokengroup import set_tokengroup, delete_tokengroup, get_tokengroups
from privacyidea.models import Tokengroup


class TokenTestCase(MyTestCase):

    def test_01_create_tokengroup(self):

        r = set_tokengroup("gruppe1", "my first typo")
        self.assertGreaterEqual(r, 1)
        tg = Tokengroup.query.filter_by(id=r).first()
        self.assertEqual(tg.Description, "my first typo")

        r = set_tokengroup("gruppe1", "my first group")
        self.assertGreaterEqual(r, 1)
        tg = Tokengroup.query.filter_by(id=r).first()
        self.assertEqual(tg.Description, "my first group")

    def test_02_delete_tokengroup(self):
        r = set_tokengroup("gruppe1", "my first group")
        self.assertGreaterEqual(r, 1)

        delete_tokengroup("gruppe1")
        tg = Tokengroup.query.filter_by(name="gruppe1").all()
        self.assertEqual(len(tg), 0)

        r = set_tokengroup("gruppe1", "my other first group")
        self.assertGreaterEqual(r, 1)

        self.assertRaises(privacyIDEAError,
                          delete_tokengroup, tokengroup_id=(r + 1), name='gruppe1')

        delete_tokengroup(tokengroup_id=r)
        tg = Tokengroup.query.filter_by(name="gruppe1").all()
        self.assertEqual(len(tg), 0)

        self.assertRaises(ResourceNotFoundError, delete_tokengroup)

    def test_03_get_tokengroups(self):
        r1 = set_tokengroup("gruppe1", "my first group")
        self.assertGreaterEqual(r1, 1)

        r2 = set_tokengroup("gruppe2", "my 2nd group")
        self.assertGreater(r2, r1)

        tgroups = get_tokengroups()
        self.assertEqual(len(tgroups), 2)

        tgroups = get_tokengroups(name="gruppe1")
        self.assertEqual(len(tgroups), 1)

        tgroups = get_tokengroups(id=r2)
        self.assertEqual(len(tgroups), 1)

        self.assertEqual(tgroups[0].name, "gruppe2")








#  2018-10-02 Friedrich Weber <friedrich.weber@netknights.it>
#             Add chunked deletions
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

from sqlalchemy import select
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import ClauseElement, Delete


class DeleteLimit(Delete, ClauseElement):
    """
    A modified DELETE clause element that allows to specify a limit
    for the number of rows that should be deleted.

    Deleting a large number of rows via just one SQL statement
    may cause deadlocks in a replicated setup. This clause can
    be used to split the one big DELETE statement into multiple
    smaller statements ("chunks") which reduces the probability
    of deadlocks.

    A dedicated clause element is needed because different
    databases use different notation for limits on DELETE statements.
    """
    inherit_cache = False

    def __init__(self, table, filter, limit=1000):
        Delete.__init__(self, table)
        self.table = table
        self.filter = filter
        if not isinstance(limit, int):
            raise RuntimeError('limit must be an integer')
        if limit <= 0:
            raise RuntimeError('limit must be positive')
        self.limit = limit


@compiles(DeleteLimit)
def visit_delete_limit(element, compiler, **kw):
    """
    Default compiler for the DeleteLimit clause element.
    This compiles to a DELETE statement with a SELECT subquery which
    has a limit set::

        DELETE FROM ... WHERE id IN
        (SELECT id FROM ... WHERE ... LIMIT ...)

    However, this syntax is not supported by MySQL.
    """
    select_stmt = select([element.table.c.id]).where(element.filter).limit(element.limit)
    delete_stmt = element.table.delete().where(element.table.c.id.in_(select_stmt))
    return compiler.process(delete_stmt)


@compiles(DeleteLimit, 'mysql')
def visit_delete_limit_mysql(element, compiler, **kw):
    """
    Special compiler for the DeleteLimit clause element
    for MySQL dialects. This compiles to a DELETE element
    with a LIMIT::

        DELETE FROM pidea_audit WHERE ... LIMIT ...
    """
    return 'DELETE FROM {} WHERE {} LIMIT {:d}'.format(  # nosec B608 # no user input used in query construction
        compiler.process(element.table, asfrom=True, **kw),
        compiler.process(element.filter), element.limit)


def delete_chunked(session, table, filter, limit=1000):
    """
    Delete all rows matching a given filter criterion from a table,
    but only delete *limit* rows at a time. Commit after each DELETE.

    :param session: SQLAlchemy session object
    :param table: SQLAlchemy table object (e.g. ``LogEntry.__table__``)
    :param filter: A filter criterion (e.g. ``LogEntry.age < now``)
    :param limit: Number of rows to delete in one chunk
    :return: total number of deleted rows
    """
    deleted = 0
    statement = DeleteLimit(table, filter, limit)
    while True:
        result = session.execute(statement)
        deleted += result.rowcount
        session.commit()
        if result.rowcount < limit:
            return deleted


def delete_matching_rows(session, table, filter, chunksize=None):
    """
    Delete all rows matching a given filter criterion from a table,
    using chunked deletes if *chunksize* is not None.

    :param session: session object
    :param table: table object
    :param filter: filter criterion
    :param chunksize: An integer (a chunksize), or None.
    :return: total number of deleted rows
    """
    if chunksize is None:
        result = session.execute(table.delete().where(filter))
        session.commit()
        return result.rowcount
    else:
        return delete_chunked(session, table, filter, chunksize)



#    2020-06-04 Cornelius Kölbel <cornelius.koelbel@netknights.i>
#               Initial Code
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU Affero General Public
#    License, version 3, as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the
#               GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

__doc__ = """This is the SMSClass to send SMS via a script.
"""

from privacyidea.lib.smsprovider.SMSProvider import (ISMSProvider, SMSError)
from privacyidea.lib import _
from privacyidea.lib.framework import get_app_config_value
import subprocess  # nosec B404 # We know what we are doing and only allow trusted scripts
import logging
import traceback
log = logging.getLogger(__name__)


SCRIPT_BACKGROUND = "background"
SCRIPT_WAIT = "wait"


class ScriptSMSProvider(ISMSProvider):

    def __init__(self, db_smsprovider_object=None, smsgateway=None, directory=None):
        """
        Create a new SMS Provider object fom a DB SMS provider object

        :param db_smsprovider_object: The database object
        :param smsgateway: The SMS gateway object from the database table
            SMS gateway. The options can be accessed via
            self.smsgateway.option_dict
        :param directory: The directory where the SMS sending scripts are located.
        :type directory: str
        :return: An SMS provider object
        """
        self.config = db_smsprovider_object or {}
        self.smsgateway = smsgateway
        self.script_directory = directory or get_app_config_value("PI_SCRIPT_SMSPROVIDER_DIRECTORY",
                                                                  "/etc/privacyidea/scripts")

    def submit_message(self, phone, message):
        """
        send a message to a phone using an external script

        :param phone: the phone number
        :param message: the message to submit to the phone
        :return:
        """
        if not self.smsgateway:
            # this should not happen. We now always use sms gateway definitions.
            log.warning("Missing smsgateway definition!")
            raise SMSError(-1, "Missing smsgateway definition!")

        phone = self._mangle_phone(phone, self.smsgateway.option_dict)
        log.debug("submitting message {0!s} to {1!s}".format(message, phone))

        script = self.smsgateway.option_dict.get("script")
        background = self.smsgateway.option_dict.get("background")

        script_name = self.script_directory + "/" + script
        proc_args = [script_name, phone]

        # As the message can contain blanks... it is passed via stdin
        rcode = 0
        try:
            log.info("Starting script {script!r}.".format(script=script_name))
            # Trusted input/no user input: The scripts are created by user root and read from hard disk
            p = subprocess.Popen(proc_args, cwd=self.script_directory,   # nosec B603
                                 universal_newlines=True, stdin=subprocess.PIPE)
            p.communicate(message)
            if background == SCRIPT_WAIT:
                rcode = p.wait()
        except Exception as e:
            log.warning("Failed to execute script {0!r}: {1!r}".format(
                script_name, e))
            log.warning(traceback.format_exc())
            if background == SCRIPT_WAIT:
                raise SMSError(-1, "Failed to start script for sending SMS.")

        if rcode:
            log.warning("Script {script!r} failed to execute with error code {error!r}".format(script=script_name,
                                                                                               error=rcode))
            raise SMSError(-1, "Error during execution of the script.")
        else:
            log.info("SMS delivered to {0!s}.".format(phone))

        return True

    @classmethod
    def parameters(cls):
        """
        Return a dictionary, that describes the parameters and options for the
        SMS provider.
        Parameters are required keys to values.

        :return: dict
        """
        params = {"options_allowed": False,
                  "parameters": {
                      "script": {
                          "required": True,
                          "description": _("The script in script directory PI_SCRIPT_SMSPROVIDER_DIRECTORY to call. "
                                           "Expects phone as the parameter and the message from stdin.")
                      },
                      "REGEXP": {
                          "description": cls.regexp_description
                      },
                      "background": {
                          "required": True,
                          "description": _("Wait for script to complete or run script in background. This will "
                                           "either return the HTTP request early or could also block the request."),
                          "values": [SCRIPT_BACKGROUND, SCRIPT_WAIT]}
                    }
                  }
        return params



#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
These are the decorators which run before processing the request and may
modify API calls e.g. by changing the flask environment.

The postAddSerialToG decorator is tested in the ValidateAPITestCase.
"""


import logging
from flask import g
import functools

log = logging.getLogger(__name__)


def add_serial_from_response_to_g(wrapped_function):
    """
    This decorator checks for the serial in the response and adds it to the
    flask g object.
    """
    @functools.wraps(wrapped_function)
    def function_wrapper(*args, **kwds):
        response = wrapped_function(*args, **kwds)
        if response.is_json:
            serial = response.json.get("detail", {}).get("serial")
            if serial:
                g.serial = serial
        return response

    return function_wrapper


# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Create/Restore database backup"""
import os
import sys
import shutil
import shlex
import pathlib
import configparser
import subprocess
import tarfile
from datetime import datetime
import re

import click
from dateutil.tz import tzlocal
from urllib.parse import urlparse
from flask import current_app
from flask.config import Config
from flask.cli import AppGroup

MYSQL_DIALECTS = ["mysql", "pymysql", "mysql+pymysql", "mariadb+pymysql"]

backup_cli = AppGroup("backup", help="Create/Restore database backup of privacyIDEA installation")


@backup_cli.command("create", short_help="Create a new backup of the database and configuration")
@click.option("-d", "--directory", "backup_dir",
              type=click.Path(file_okay=False, writable=True),
              default="/var/lib/privacyidea/backup/",
              show_default=True,
              help="Path to the backup directory")
@click.option("-c", "--config_dir",
              type=click.Path(exists=True, file_okay=False, writable=True),
              default="/etc/privacyidea/",
              show_default=True,
              help="Path to privacyIDEA config directory")
@click.option("-r", "--radius_dir",
              type=click.Path(exists=True, file_okay=False, readable=True),
              default=None,
              show_default=True,
              help="Path to FreeRADIUS config directory")
@click.option("-e", "--enckey", is_flag=True,
              help="Add the encryption key to the backup")
def backup_create(backup_dir, config_dir, radius_dir, enckey):
    """
    Create a new backup of the database and the configuration. By default,
    the encryption key is not included. Use the 'enckey' option to also
    add the encryption key to the backup. In this case make sure, that the
    backups are stored securely.

    You can also include a given FreeRADIUS configuration into the backup.
    Just specify a directory using 'radius_dir'.
    """
    # TODO: Add requirement for the config file and remove app initialization.
    #  Currently, when calling this function, the Flask app gets initialized
    #  (either from /etc/privacyidea/pi.cfg or from the environment variable)
    #  regardless of the given config directory (so they can differ).
    #  Since all the necessary paths are given in the config file, we should
    #  just use that for the gathering the files for backup.
    # TODO: Remove generated/copied file in case of an error. Maybe create a
    #  temporary folder where the data is collected
    cur_date = datetime.now(tz=tzlocal()).strftime("%Y%m%d-%H%M")
    base_name = "privacyidea-backup"

    conf_dir = pathlib.Path(config_dir).absolute()
    directory = pathlib.Path(backup_dir).absolute()
    directory.mkdir(parents=True, exist_ok=True)

    enc_file = pathlib.Path(current_app.config.get("PI_ENCFILE"))

    # set correct owner, if possible
    if os.geteuid() == 0:
        enc_file_stat = enc_file.stat()
        shutil.chown(directory, user=enc_file_stat.st_uid, group=enc_file_stat.st_gid)

    sqlfile = directory.joinpath(f"dbdump-{cur_date}.sql")
    backup_file = directory.joinpath(f"{base_name}-{cur_date}.tgz")

    parsed_sqluri = urlparse(current_app.config.get("SQLALCHEMY_DATABASE_URI"))
    sqltype = parsed_sqluri.scheme

    if sqltype == "sqlite":
        productive_file = parsed_sqluri.path
        click.echo(f"Backup SQLite file {productive_file}")
        sqlfile = directory.joinpath(f"dbdump-{cur_date}.sqlite")
        shutil.copyfile(productive_file, sqlfile)
    elif sqltype in MYSQL_DIALECTS:
        database = parsed_sqluri.path[1:]
        defaults_file = conf_dir.joinpath("mysql.cnf")
        _write_mysql_defaults(defaults_file, parsed_sqluri)
        # call mysqldump to get a copy of the database
        cmd = ['mysqldump', '--defaults-file={!s}'.format(defaults_file), '-h',
               shlex.quote(parsed_sqluri.hostname)]
        if parsed_sqluri.port:
            cmd.extend(['-P', str(parsed_sqluri.port)])
        cmd.extend(['-B', shlex.quote(database), '-r', sqlfile])
        subprocess.run(cmd)
    else:
        click.echo(f"unsupported SQL syntax: {sqltype}")
        sys.exit(2)

    with tarfile.open(backup_file, "x:gz") as tf:
        tf.add(sqlfile)
        if radius_dir:
            # Simply append the radius directory to the backup command
            tf.add(radius_dir, recursive=True)

        if not enckey:
            # Exclude enckey from backup

            def exclude_encfile(t):
                if enc_file.match(t.name):
                    return None
                return t

            tf.add(conf_dir, recursive=True, filter=exclude_encfile)
        else:
            click.secho("Including encryption key in backup", fg="yellow")
            tf.add(conf_dir, recursive=True)

    sqlfile.unlink()
    backup_file.chmod(0o600)
    click.echo(f"Backup written to file {backup_file}")


@backup_cli.command("restore")
@click.argument("backup_file", type=str)
def backup_restore(backup_file):
    """Restore a previously made backup from the BACKUP_FILE"""
    # TODO: Use tarfile package
    # TODO: Also allow to specify a target directory, otherwise it will always
    #  extract to the base /
    # TODO: extracting the SQLite file does not work if there are other SQLite
    #  files in the archive
    sqluri = None
    config_file = None
    sqlfile = None
    enckey_contained = False

    p = subprocess.run(["tar", "-ztf", backup_file], capture_output=True,
                       text=True)
    if p.returncode != 0:
        click.secho(f"Unable to open backup file {backup_file}", fg="red")
        sys.exit(2)
    for line in p.stdout.split("\n"):
        if re.search(r"/pi.cfg$", line):
            config_file = "/{0!s}".format(line.strip())
        elif re.search(r"dbdump-\d{8}-\d{4}\.sql", line):
            sqlfile = "/{0!s}".format(line.strip())
        elif re.search(r"/enc[kK]ey", line):
            enckey_contained = True

    if not config_file:
        click.secho("Missing config file pi.cfg in backup file.", fg="red")
    if not sqlfile:
        click.secho("Missing database dump in backup file.", fg="red")

    config_file = pathlib.Path(config_file)
    sqlfile = pathlib.Path(sqlfile)
    if enckey_contained:
        click.echo("Also restoring the encryption key")
    else:
        click.secho("NO FILE 'enckey' CONTAINED! BE SURE TO RESTORE THE ENCRYPTION "
                    "KEY MANUALLY!", fg='yellow')
    click.echo(f"Restoring to {config_file} with data from {sqlfile}")

    subprocess.run(["tar", "-zxf", backup_file, "-C", "/"])
    click.echo(60 * "=")
    # use Flask config to read in the config file
    cfg = Config(config_file.parent)
    cfg.from_pyfile(config_file)
    sqluri = cfg["SQLALCHEMY_DATABASE_URI"]

    if sqluri is None:
        click.secho(f"No SQLALCHEMY_DATABASE_URI found in {config_file}",
                    fg="red")
        sys.exit(2)
    parsed_sqluri = urlparse(sqluri)
    sqltype = parsed_sqluri.scheme
    if sqltype == "sqlite":
        productive_file = parsed_sqluri.path
        click.echo(f"Restore SQLite {productive_file}")
        shutil.copyfile(sqlfile, productive_file)
        os.unlink(sqlfile)
    elif sqltype in MYSQL_DIALECTS:
        database = parsed_sqluri.path[1:]
        defaults_file = pathlib.Path(config_file).parent.joinpath("mysql.cnf")
        _write_mysql_defaults(defaults_file, parsed_sqluri)
        # Rewriting database
        click.echo("Restoring database.")
        cmd = ["mysql", f"--defaults-file={defaults_file}",
               "-h", parsed_sqluri.hostname]
        if parsed_sqluri.port:
            cmd.extend(['-P', str(parsed_sqluri.port)])
        cmd.extend(['-B', shlex.quote(database)])
        with open(sqlfile, "r") as sql_file:
            p = subprocess.run(cmd, input=sql_file.read(), text=True)
            if p.returncode == 0:
                os.unlink(sqlfile)
    else:
        print("unsupported SQL syntax: %s" % sqltype)
        sys.exit(2)


def _write_mysql_defaults(defaults_file, parsed_sqluri):
    # create a mysql config file to avoid adding username and password to the command
    sql_defaults = configparser.ConfigParser()
    sql_defaults['client'] = {
        "user": parsed_sqluri.username,
        "password": parsed_sqluri.password
    }
    sql_defaults['mysqldump'] = {"no-tablespaces": "True"}
    with defaults_file.open(mode="w") as f:
        sql_defaults.write(f)
    defaults_file.chmod(0o600)
    # set correct owner, if possible
    if os.geteuid() == 0:
        dir_stat = defaults_file.parent.stat()
        shutil.chown(defaults_file, dir_stat.st_uid, dir_stat.st_gid)



# 2024-03-22 Jona-Samuel Höhmann <jona-samuel.hoehmann@netknights.it>
#            Migrate to click
# 2017-04-24 Friedrich Weber <friedrich.weber@netknights.it>
#
# Copyright (c) 2017, Friedrich Weber
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
This script deletes expired entries from the user cache.
"""
__version__ = "0.1"

from flask.cli import with_appcontext, ScriptInfo
from privacyidea.lib.usercache import create_filter, get_cache_time
from privacyidea.lib.utils import get_version_number
from privacyidea.models import UserCache, db
from privacyidea.cli import create_silent_app
import click

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def _get_expired_entries():
    """
    Returns a list of all cache entries that are considered expired.
    """
    filter_condition = create_filter(expired=True)
    return UserCache.query.filter(filter_condition).order_by(UserCache.timestamp.desc()).all()


LIST_FORMAT = '{:<5} {:<10} {:<10} {:<30} {:<10}'


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-n', "--noaction", is_flag=True, default=False)
@with_appcontext
def delete(noaction=False):
    """
    Delete all cache entries that are considered expired according to the
    UserCacheExpiration configuration setting.
    If the user cache is disabled, no entries are deleted.
    If '--noaction' is passed, expired cache entries are listed, but not
    actually deleted.
    """
    if not get_cache_time():
        click.echo('User cache is disabled, not doing anything.')
    else:
        click.echo('Expired entries:')
        entries = _get_expired_entries()
        if entries:
            click.echo(LIST_FORMAT.format('id', 'username', 'resolver', 'timestamp', 'user id'))
        for entry in entries:
            click.echo(LIST_FORMAT.format(entry.id,
                                          entry.username,
                                          entry.resolver,
                                          entry.timestamp.isoformat(),
                                          entry.user_id))
        click.echo('{} entries'.format(len(entries)))
        if not noaction:
            for entry in entries:
                click.echo('Deleting entry with id={} ...'.format(entry.id))
                entry.delete()
            click.echo('Deleted {} expired entries.'.format(len(entries)))
            db.session.commit()
        else:
            click.echo("'--noaction' was passed, not doing anything.")


def delete_call():
    """Management script for usercache cleanup of privacyIDEA."""
    click.echo(r"""
                 _                    _______  _______
       ___  ____(_)  _____ _______ __/  _/ _ \/ __/ _ |
      / _ \/ __/ / |/ / _ `/ __/ // // // // / _// __ |
     / .__/_/ /_/|___/\_,_/\__/\_, /___/____/___/_/ |_|  Usercache cleanup
    /_/                       /___/
    {0!s:>51}
        """.format('v{0!s}'.format(get_version_number())))

    # Add the ScriptInfo object to create the Flask-App when necessary
    s = ScriptInfo(create_app=create_silent_app)
    delete(obj=s)


if __name__ == '__main__':
    delete_call()



# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import datetime
import click
from flask import current_app
from flask.cli import AppGroup
import jwt

from privacyidea.lib.auth import ROLE
from privacyidea.lib.crypto import geturandom


api_cli = AppGroup("api", help="Manage API keys")


@api_cli.command("createtoken")
@click.option('-r', '--role',
              help="The role of the API key can either be "
                   "'admin' or 'validate' to access the admin "
                   "API or the validate API.",
              type=click.Choice([ROLE.ADMIN, ROLE.VALIDATE]),
              default=ROLE.ADMIN, show_default=True)
@click.option('-d', '--days',
              help='The number of days the access token should be valid.',
              default=365, show_default=True, type=int)
@click.option('-R', '--realm',
              help='The realm of the admin.',
              default="API", type=str, show_default=True)
@click.option('-u', '--username', type=str, required=True,
              help='The username of the admin.')
@click.pass_context
def api_createtoken(ctx, role, days, realm, username):
    """
    Create an API authentication token
    for administrative or validate use.
    Possible roles are "admin" or "validate".
    """
    if role not in ["admin", "validate"]:
        click.secho("ERROR: The role must be 'admin' or 'validate'!", fg="red")
        ctx.exit(1)
    username = username or geturandom(hex=True)
    secret = current_app.config.get("SECRET_KEY")
    authtype = "API"
    validity = datetime.timedelta(days=int(days))
    token = jwt.encode({
        "username": username,
        "realm": realm,
        "nonce": geturandom(hex=True),
        "role": role,
        "authtype": authtype,
        "exp": datetime.datetime.utcnow() + validity,
        "rights": "TODO"},
        secret)
    click.echo(f"Username:   {username}")
    click.echo(f"Realm:      {realm}")
    click.echo(f"Role:       {role}")
    click.echo(f"Validity:   {days} days")
    click.echo(f"Auth-Token: {token}")



""" API testcases for the "/system/ endpoint """
import gnupg
import json
import os
import unittest
from urllib.parse import urlencode

from .base import MyApiTestCase

from privacyidea.lib.policy import PolicyClass, set_policy, delete_policy, ACTION, SCOPE
from privacyidea.lib.caconnector import save_caconnector, delete_caconnector
from privacyidea.lib.caconnectors.localca import ATTR
from privacyidea.lib.radiusserver import add_radius, delete_radius
from privacyidea.lib.resolver import save_resolver, delete_resolver, CENSORED
from privacyidea.lib.realm import delete_realm, get_realms
from privacyidea.models import db, NodeName
from .test_lib_resolver import LDAPDirectory, ldap3mock
from .test_lib_caconnector import CACERT, CAKEY, WORKINGDIR, OPENSSLCNF

PWFILE = "tests/testdata/passwords"
POLICYFILE = "tests/testdata/policy.cfg"
POLICYEMPTY = "tests/testdata/policy_empty_file.cfg"

try:
    _g = gnupg.GPG()
    gpg_available = True
except OSError as _e:
    gpg_available = False


class APIConfigTestCase(MyApiTestCase):

    def test_00_get_empty_config(self):
        with self.app.test_request_context('/system/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            self.assertTrue(res.json['result']['status'], res.json)

    def test_00_failed_auth(self):
        with self.app.test_request_context('/system/',
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

    def test_01_set_config(self):
        with self.app.test_request_context('/system/setConfig',
                                           data={"key1": "value1",
                                                 "key2": "value2",
                                                 "key3": "value3"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            value = res.json['result']['value']
            self.assertEqual(value.get("key1"),
                             "insert",
                             "This sometimes happens when test database was "
                             "not empty!")
            self.assertEqual(value.get("key2"), "insert")
            self.assertEqual(value.get("key3"), "insert")

    def test_02_update_config(self):
        with self.app.test_request_context('/system/setConfig',
                                           data={"key3": "new value"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            self.assertEqual(res.json['result']['value']['key3'], 'update',
                             res.json)

    def test_03_set_and_del_default(self):
        with self.app.test_request_context('/system/setDefault',
                                           data={"DefaultMaxFailCount": 1,
                                                 "DefaultSyncWindow": 10,
                                                 "DefaultCountWindow": 12,
                                                 "DefaultOtpLen": 6,
                                                 "DefaultResetFailCount": 12},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            self.assertTrue(result["status"], result)
            self.assertTrue(result["value"]["DefaultOtpLen"] == "insert",
                            result)

        with self.app.test_request_context('/system/DefaultMaxFailCount',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(result["value"], 1, result)

        with self.app.test_request_context('/system/DefaultMaxFailCount',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertIsNone(result["value"], result)

        # test unknown parameter
        with self.app.test_request_context('/system/setDefault',
                                           data={"unknown": "xx"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            # "unknown" is an unknown Default Parameter. So a ParamterError
            # is raised.
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

    def test_04_set_policy(self):
        with self.app.test_request_context('/policy/pol1',
                                           data={'action': "enroll",
                                                 'scope': "selfservice",
                                                 'realm': "r1",
                                                 'resolver': "test",
                                                 'user': ["admin"],
                                                 'time': "",
                                                 'client': "127.12.12.12",
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(result['value']['setPolicy pol1'], 1, res.json)

        # Update the policy with a more complicated client
        with self.app.test_request_context('/policy/pol1',
                                           data={'action': "enroll",
                                                 'scope': "selfservice",
                                                 'realm': "r1",
                                                 'resolver': "test",
                                                 'user': ["admin"],
                                                 'time': "",
                                                 'client': "10.0.0.0/8, "
                                                           "172.16.200.1",
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json['result']
            self.assertTrue(result["status"], result)
            self.assertGreaterEqual(result['value']['setPolicy pol1'], 1, result)

        # setting policy with invalid name fails
        with self.app.test_request_context('/policy/invalid policy name',
                                           data={'action': "enroll",
                                                 'scope': "selfservice",
                                                 'client': "127.12.12.12",
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            # An invalid policy name raises an exception
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

        # setting policy with an empty name
        with self.app.test_request_context('/policy/enroll',
                                           data={'scope': "selfservice",
                                                 'client': "127.12.12.12",
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            # An invalid policy name raises an exception
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

    def test_05_get_policy(self):
        with self.app.test_request_context('/policy/pol1',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue("pol1" == result["value"][0].get("name"), res.data)

    def test_06_export_policy(self):
        with self.app.test_request_context('/policy/export/test.cfg',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            body = res.data
            self.assertTrue(b'name = pol1' in body, res.data)
            self.assertTrue(b"[pol1]" in body, res.data)
        delete_policy("pol1")

    def test_07_update_and_delete_policy(self):
        with self.app.test_request_context('/policy/pol_update_del',
                                           data={'action': "enroll",
                                                 'scope': "selfservice",
                                                 'realm': "r1",
                                                 'resolver': "test",
                                                 'user': "admin",
                                                 'time': "",
                                                 'client': "127.12.12.12",
                                                 'active': True},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue(result["value"]["setPolicy pol_update_del"] > 0,
                            res.data)

        # update policy
        with self.app.test_request_context('/policy/pol_update_del',
                                           data={'action': "enroll",
                                                 'scope': "selfservice",
                                                 'realm': "r1",
                                                 'client': "1.1.1.1"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["value"]["setPolicy pol_update_del"] > 0,
                            res.data)

        # get policy
        with self.app.test_request_context('/policy/pol_update_del',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            policy = {}
            for pol in result["value"]:
                if pol.get("name") == "pol_update_del":
                    policy = pol
                    break
            self.assertTrue("1.1.1.1" in policy.get("client"),
                            res.data)

        # delete policy again does not do anything
        with self.app.test_request_context('/policy/pol_update_del',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)

        # delete policy
        with self.app.test_request_context('/policy/pol_update_del',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404)
            result = res.json.get("result")
            self.assertFalse(result["status"])

        # check policy
        with self.app.test_request_context('/policy/pol_update_del',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue(result["value"] == [], result)

    # Resolvers
    """
    We should move this to LDAP resolver tests and mock this.

    def test_08_pretestresolver(self):
        # This test fails, as there is no server at localhost.
        param = {'LDAPURI': 'ldap://localhost',
                 'LDAPBASE': 'o=test',
                 'BINDDN': 'cn=manager,ou=example,o=test',
                 'BINDPW': 'ldaptest',
                 'LOGINNAMEATTRIBUTE': 'cn',
                 'LDAPSEARCHFILTER': '(cn=*)',
                 'USERINFO': '{ "username": "cn",'
                             '"phone" : "telephoneNumber", '
                             '"mobile" : "mobile"'
                             ', "email" : "mail", '
                             '"surname" : "sn", '
                             '"givenname" : "givenName" }',
                 'UIDTYPE': 'DN',
                 'type': 'ldapresolver'}
        with self.app.test_request_context('/resolver/test',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertFalse(result.get("value"), result)
            self.assertTrue("no active server available in server pool" in
                            detail.get("description"),
                            detail.get("description"))
    """
    def test_08_no_ldap_password(self):
        params = {'LDAPURI': 'ldap://localhost',
                  'LDAPBASE': 'o=test',
                  'BINDDN': 'cn=manager,ou=example,o=test',
                  'BINDPW': 'ldaptest',
                  'LOGINNAMEATTRIBUTE': 'cn',
                  'LDAPSEARCHFILTER': '(cn=*)',
                  'USERINFO': '{ "username": "cn", "phone": "telephoneNumber", '
                              '"mobile" : "mobile", "email": "mail", '
                              '"surname" : "sn", "givenname": "givenName" }',
                  'UIDTYPE': 'DN',
                  'type': 'ldapresolver',
                  'resolver': 'testL'}
        r = save_resolver(params)
        self.assertTrue(r)
        with self.app.test_request_context('/resolver/testL',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            data = result["value"]["testL"]["data"]
            self.assertEqual(data.get("BINDPW"), CENSORED)

        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            data = result["value"]["testL"]["data"]
            self.assertEqual(data.get("BINDPW"), CENSORED)

        r = delete_resolver(params.get("resolver"))
        self.assertTrue(r)

    def test_08_resolvers(self):
        with self.app.test_request_context('/resolver/resolver1',
                                           data={'type': 'passwdresolver',
                                                 'filename': PWFILE},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertGreaterEqual(result["value"], 1, result)
            res_id = result["value"]

        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue("resolver1" in result["value"], result)
            self.assertTrue("filename" in result["value"]["resolver1"]["data"])

        # Get a non existing resolver
        with self.app.test_request_context('/resolver/unknown',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The value is empty
            self.assertTrue(result["value"] == {}, result)

        # Get only editable resolvers
        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           query_string=urlencode({
                                               "editable": "1"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The value is empty
            self.assertTrue(result["value"] == {}, result)

        # this will fetch all resolvers
        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            self.assertTrue("resolver1" in value, value)

        # get non-editable resolvers
        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           query_string=urlencode({
                                               "editable": "0"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            self.assertTrue("resolver1" in value, value)

        # get a resolver name
        with self.app.test_request_context('/resolver/resolver1',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            self.assertTrue(result["status"], result)
            self.assertTrue("resolver1" in result["value"], result)
            self.assertTrue("filename" in result["value"]["resolver1"]["data"])

        # get a resolver name
        with self.app.test_request_context('/resolver/resolver1',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue("resolver1" in result["value"], result)
            self.assertTrue("filename" in result["value"]["resolver1"]["data"])

        # delete the resolver
        with self.app.test_request_context('/resolver/resolver1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            self.assertTrue(result["status"], result)
            self.assertEqual(result["value"], res_id, result)

        # delete a non existing resolver
        with self.app.test_request_context('/resolver/xycswwf',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            self.assertTrue(result["status"], result)
            # Trying to delete a non-existing resolver returns -1
            self.assertTrue(result["value"] == -1, result)

    def test_09_handle_realms(self):
        resolvername = "reso1_with_realm"
        realmname = "realm1_with_resolver"
        # create a resolver
        with self.app.test_request_context('/resolver/{0!s}'.format(resolvername),
                                           method='POST',
                                           data={"filename": PWFILE,
                                                 "type": "passwdresolver"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The resolver was created. The ID of the resolver is returend.
            self.assertGreaterEqual(result["value"], 1, result)
            res_id = result["value"]

        # create a realm
        with self.app.test_request_context('/realm/{0!s}'.format(realmname),
                                           method='POST',
                                           data={"resolvers": resolvername},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The resolver was created
            self.assertTrue(len(result["value"].get("added")) == 1, result)
            self.assertTrue(len(result["value"].get("failed")) == 0, result)

        # create a realm with multiple resolvers
        with self.app.test_request_context('/realm/realm2',
                                           method='POST',
                                           json={"resolvers": [resolvername, "resolver2"]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The resolver was created
            self.assertEqual(len(result["value"].get("added")), 1, result)
            self.assertEqual(len(result["value"].get("failed")), 1, result)

        # display the realm
        with self.app.test_request_context('/realm/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The resolver was created = 1
            self.assertTrue(realmname in result["value"], result)
            realm_contents = result["value"].get(realmname)
            self.assertTrue(realm_contents.get("resolver")[0].get("name") ==
                            resolvername, result)

        # get the superuser realms
        with self.app.test_request_context('/realm/superuser',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue("adminrealm" in result["value"], result)

        # try to delete the resolver in the realm
        with self.app.test_request_context('/resolver/{0!s}'.format(resolvername),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            # The resolver must not be deleted, since it is contained in a realm
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

        # delete the realm
        with self.app.test_request_context('/realm/{0!s}'.format(realmname),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            # The realm gets deleted
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The realm is successfully deleted: value is the id in
            # the db, should be >= 1
            self.assertGreaterEqual(result["value"], 1, result)

        # delete the second realm
        with self.app.test_request_context('/realm/realm2',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            # The realm gets deleted
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The realm is successfully deleted: value is the id in
            # the db, should be >= 1
            self.assertGreaterEqual(result["value"], 1, result)

        # Now, we can delete the resolver
        with self.app.test_request_context('/resolver/{0!s}'.format(resolvername),
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            # The resolver must not be deleted, since it is contained in a realm
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            # The resolver was deleted = 1
            self.assertEqual(result["value"], res_id, result)

    def test_10_default_realm(self):
        resolvername = "defresolver"
        realmname = "defrealm"
        with self.app.test_request_context('/resolver/{0!s}'.format(resolvername),
                                           method='POST',
                                           data={"filename": PWFILE,
                                                 "type": "passwdresolver"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)

        # create a realm
        with self.app.test_request_context('/realm/{0!s}'.format(realmname),
                                           method='POST',
                                           data={"resolvers": resolvername,
                                                 "priority.defresolver": 10},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)

        # get the default realm
        with self.app.test_request_context('/defaultrealm',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue("defrealm" in result["value"], result)

        # clear the default realm
        with self.app.test_request_context('/defaultrealm',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)

        # get the default realm
        with self.app.test_request_context('/defaultrealm',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue(result["value"] == {}, result)

        # set the default realm
        with self.app.test_request_context('/defaultrealm/defrealm',
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)

        # get the default realm
        with self.app.test_request_context('/defaultrealm',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            self.assertEqual(res.status_code, 200, res)
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertTrue("defrealm" in result["value"], result)

    def test_11_import_policy(self):
        with self.app.test_request_context('/policy/import/policy.cfg',
                                           method='POST',
                                           data=dict(file=(POLICYFILE,
                                                           'policy.cfg')),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(result["value"], 2, result)
            # check if policies are there
            pol = PolicyClass()
            p1 = pol.match_policies(name="importpol1")
            self.assertTrue(len(p1) == 1, p1)
            p2 = pol.match_policies(name="importpol2")
            self.assertTrue(len(p2) == 1, p2)

        # import empty file
        with self.app.test_request_context("/policy/import/"
                                           "policy_empty_file.cfg",
                                           method='POST',
                                           data=dict(file=(POLICYEMPTY,
                                                           "policy_empty_file.cfg")),
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

    def test_12_test_check_policy(self):
        # test invalid policy name "check"
        with self.app.test_request_context('/policy/check',
                                           method='POST',
                                           data={"realm": "*",
                                                 "action": "action1, action2",
                                                 "scope": "scope1",
                                                 "user": "*, -user1",
                                                 "client": "172.16.0.0/16, "
                                                           "-172.16.1.1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

        with self.app.test_request_context('/policy/pol1',
                                           method='POST',
                                           data={"realm": "*",
                                                 "action": "action1, action2",
                                                 "scope": "scope1",
                                                 "user": "*, -user1",
                                                 "client": "172.16.0.0/16, "
                                                           "-172.16.1.1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertGreater(result.get("value")["setPolicy pol1"], 0, result)

        with self.app.test_request_context('/policy/pol2',
                                           method='POST',
                                           data={"realm": "*",
                                                 "action": "action3=value, "
                                                           "action4",
                                                 "scope": "scope1",
                                                 "user": "admin, superuser",
                                                 "client": "172.16.1.1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertGreater(result.get("value")["setPolicy pol2"], 1, result)

        # CHECK: user=superuser, action=action1, client=172.16.1.1
        # is not allowed
        with self.app.test_request_context('/policy/check',
                                           method='GET',
                                           query_string=urlencode({"realm": "realm1",
                                                                   "action": "action1",
                                                                   "scope": "scope1",
                                                                   "user": "superuser",
                                                                   "client": "172.16.1.1"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertFalse(result.get("value").get("allowed"), result)

        # CHECK: user=superuser, action=action1, client=172.16.1.2
        # is allowed
        with self.app.test_request_context('/policy/check',
                                           method='GET',
                                           query_string=urlencode({"realm": "realm2",
                                                                   "action": "action1",
                                                                   "scope": "scope1",
                                                                   "user": "superuser",
                                                                   "client": "172.16.1.2"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value").get("allowed"), result)

        # CHECK: user=superuser, action=action3, client=172.16.1.2
        # is not allowed
        with self.app.test_request_context('/policy/check',
                                           method='GET',
                                           query_string=urlencode({"realm": "realm3",
                                                                   "action": "action3",
                                                                   "scope": "scope1",
                                                                   "user": "superuser",
                                                                   "client": "172.16.1.2"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertFalse(result.get("value").get("allowed"), result)

        # CHECK: user=superuser, action=action3, client=172.16.1.1
        # is allowed
        with self.app.test_request_context('/policy/check',
                                           method='GET',
                                           query_string=urlencode({"realm": "realm1",
                                                                   "action": "action3",
                                                                   "scope": "scope1",
                                                                   "user": "superuser",
                                                                   "client": "172.16.1.1"}),
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value").get("allowed"), result)

    def test_13_get_policy_defs(self):
        with self.app.test_request_context('/policy/defs',
                                           method='GET',
                                           data={},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            policies = result.get("value")
            admin_pol = policies.get("admin")
            self.assertTrue("enable" in admin_pol, admin_pol)
            self.assertTrue("enrollTOTP" in admin_pol, admin_pol)
            self.assertTrue("enrollHOTP" in admin_pol, admin_pol)
            self.assertTrue("enrollPW" in admin_pol, admin_pol)

        with self.app.test_request_context('/policy/defs/admin',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            admin_pol = result.get("value")
            self.assertTrue("enable" in admin_pol, admin_pol)
            self.assertTrue("enrollTOTP" in admin_pol, admin_pol)
            self.assertTrue("enrollHOTP" in admin_pol, admin_pol)
            self.assertTrue("enrollPW" in admin_pol, admin_pol)

        with self.app.test_request_context('/policy/defs/conditions',
                                           method='GET',
                                           data={},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            conditions = result.get("value")
            self.assertIn("sections", conditions)
            self.assertIn("userinfo", conditions["sections"])
            self.assertIn("description", conditions["sections"]["userinfo"])
            self.assertIn("comparators", conditions)
            self.assertIn("contains", conditions["comparators"])
            self.assertIn("description", conditions["comparators"]["contains"])

    def test_14_enable_disable_policy(self):
        with self.app.test_request_context('/policy/pol2',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            pol = result.get("value")
            self.assertTrue(pol[0].get("active"), pol[0])

        # Disable policy
        with self.app.test_request_context('/policy/disable/pol2',
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertGreater(result.get("value"), 0, result)

        with self.app.test_request_context('/policy/pol2',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            pol = result.get("value")
            self.assertFalse(pol[0].get("active"), pol[0])

        # enable Policy
        with self.app.test_request_context('/policy/enable/pol2',
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            self.assertTrue(result.get("status"), result)
            self.assertGreater(result.get("value"), 0, result)

        with self.app.test_request_context('/policy/pol2',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertEqual(res.status_code, 200, res)
            pol = result.get("value")
            self.assertTrue(pol[0].get("active"), pol[0])

    def test_15_get_documentation(self):
        with self.app.test_request_context('/system/documentation',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            self.assertEqual(res.mimetype, 'text/plain', res)
            self.assertTrue(b"privacyIDEA configuration documentation" in
                            res.data)

    def test_16_get_hsm(self):
        with self.app.test_request_context('/system/hsm',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            self.assertTrue(value.get("is_ready"), value)

        # HSM is already set up. We do not need to set a password
        with self.app.test_request_context('/system/hsm',
                                           data={"password": "xx"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)

    def test_17_test_token_config(self):
        with self.app.test_request_context('/system/test/hotp',
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            details = res.json.get("detail")
            value = result.get("value")
            self.assertEqual(value, False)
            self.assertEqual(details.get("message"), "Not implemented")

    def test_18_test_random(self):
        with self.app.test_request_context('/system/random?len=32',
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            # hex encoded value
            self.assertEqual(len(value), 64)
            # This is hex, we can unhexlify
            import binascii
            binascii.unhexlify(value)

        with self.app.test_request_context('/system/random?len=32&encode=b64',
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            # hex encoded value
            self.assertEqual(len(value), 44)
            # This is base64. We can decode
            import base64
            base64.b64decode(value)

    @unittest.skipIf(not gpg_available, "'gpg' binary not available")
    def test_19_get_gpg_keys(self):
        with self.app.test_request_context('/system/gpgkeys',
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            value = result.get("value")
            # We probably have no keys in here
            # But value returns a dictionary with the KeyID and "armor" and
            # "fingerprint"
            self.assertIn("2F25BAF8645350BB", value, value)
            key_obj = value.get("2F25BAF8645350BB")
            self.assertIn("-----BEGIN PGP PUBLIC KEY BLOCK-----", key_obj.get("armor"), key_obj)
            self.assertEqual("6630FE8C6866433020D39FA02F25BAF8645350BB", key_obj.get("fingerprint"), value)

    @ldap3mock.activate
    def test_20_multiple_test_resolvers(self):
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        params = {'LDAPURI': 'ldap://localhost',
                  'LDAPBASE': 'o=test',
                  'BINDDN': 'cn=manager,ou=example,o=test',
                  'BINDPW': 'ldaptest',
                  'LOGINNAMEATTRIBUTE': 'cn',
                  'LDAPSEARCHFILTER': '(cn=*)',
                  'USERINFO': '{ "username": "cn", "phone": "telephoneNumber", '
                              '"mobile" : "mobile", "email": "mail", '
                              '"surname" : "sn", "givenname": "givenName" }',
                  'UIDTYPE': 'DN'}

        resolvername = "blablaReso"
        params["resolver"] = resolvername
        params["type"] = "ldapresolver"
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           data=params,
                                           method="POST",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertGreater(result["value"], 0, result)

        with (((self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                              method="GET",
                                              headers={'Authorization': self.at})))):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            params = result.get("value").get(resolvername).get("data")
            # the returned password is censored
            self.assertEqual(params.get("BINDPW"), CENSORED)
            # the intenal password is correct
            int_res_cnf = self.app_context.g._request_local_store.get("config_object").resolver.get(resolvername)
            self.assertEqual(int_res_cnf.get("data").get("BINDPW"), "ldaptest")

    def test_21_read_write_resolver_policy(self):
        # create resolver
        resolvername = "reso21"
        params = {"resolver": resolvername,
                  "type": "passwdresolver",
                  "file": "/etc/passwd"}
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           data=params,
                                           method="POST",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # Set a read policy
        set_policy(name="pol_read", scope=SCOPE.ADMIN,
                   action=ACTION.RESOLVERREAD)

        # Now writing a resolver will fail
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           data=params,
                                           method="POST",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)

        # reading a resolver will succeed
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # set a write policy
        set_policy(name="pol_write", scope=SCOPE.ADMIN,
                   action=ACTION.RESOLVERWRITE)

        # Now writing a resolver will succeed
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           data=params,
                                           method="POST",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # delete read policy
        delete_policy("pol_read")

        # reading a resolver will fail
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)

        # writing a resolver will still succeed
        with self.app.test_request_context('/resolver/{0}'.format(resolvername),
                                           data=params,
                                           method="POST",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        delete_policy("pol_write")
        delete_resolver(resolvername)

    def test_22_list_radius_servers(self):
        add_radius("local", "localhost", "foobar")
        add_radius("remote", "remote.example.com", "secret", description="very far away")

        # cannot access the RADIUS server names without authorization
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertIn("Missing Authorization header", result["error"]["message"])

        # if no admin policies are defined, admins can access the RADIUS servers
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertEqual(set(result["value"]), {"local", "remote"})

        # if an admin policy is defined and enrollRADIUS is not allowed, admins cannot access the RADIUS servers
        set_policy("admin", scope=SCOPE.ADMIN, action=ACTION.AUDIT)
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertIn("enrollRADIUS", result["error"]["message"])

        # with an enrollRADIUS action, admins can access the RADIUS servers
        set_policy("admin", scope=SCOPE.ADMIN, action=[ACTION.AUDIT, "enrollRADIUS"])
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertEqual(set(result["value"]), {"local", "remote"})

        self.setUp_user_realms()
        self.authenticate_selfservice_user()

        # without a user action, users can access the RADIUS servers
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertEqual(set(result["value"]), {"local", "remote"})

        # if a user policy is defined and enrollRADIUS is not allowed, users cannot access the RADIUS servers
        set_policy("user", scope=SCOPE.USER, action=ACTION.AUDIT)
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertIn("enrollRADIUS", result["error"]["message"])

        # with an enrollRADIUS action, users can access the RADIUS servers
        set_policy("user", scope=SCOPE.USER, action="enrollRADIUS")
        with self.app.test_request_context("/system/names/radius",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertEqual(set(result["value"]), {"local", "remote"})

        delete_policy("user")
        delete_policy("admin")
        delete_radius("local")
        delete_radius("remote")

    def test_23_list_ca_connectors(self):
        cwd = os.getcwd()
        save_caconnector({'type': 'local',
                          'secret': 'value',
                          'caconnector': 'localCA',
                          "cakey": CAKEY,
                          "cacert": CACERT,
                          "openssl.cnf": OPENSSLCNF,
                          "WorkingDir": cwd + "/" + WORKINGDIR,
                          ATTR.TEMPLATE_FILE: "templates.yaml"})

        def _check_caconnector_response(response):
            value = json.loads(response.data.decode('utf8')).get("result")["value"]
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0]["connectorname"], "localCA")
            self.assertEqual(value[0]["data"], {})
            self.assertEqual(set(value[0]["templates"].keys()), {"template3", "webserver", "user"})

        # cannot access the CA connector names without authorization
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertIn("Missing Authorization header", result["error"]["message"])

        # if no admin policies are defined, admins can access the CA connectors
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            _check_caconnector_response(res)

        # if an admin policy is defined and enrollCERTIFICATE is not allowed, admins cannot access the CA connectors
        set_policy("admin", scope=SCOPE.ADMIN, action=ACTION.AUDIT)
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertIn("enrollCERTIFICATE", result["error"]["message"])

        # with an enrollCERTIFICATE action, admins can access the CA connectors
        set_policy("admin", scope=SCOPE.ADMIN, action=[ACTION.AUDIT, "enrollCERTIFICATE"])
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            _check_caconnector_response(res)

        self.setUp_user_realms()
        self.authenticate_selfservice_user()

        # without a user action, users can access the CA connectors
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            _check_caconnector_response(res)

        # if a user policy is defined and enrollCERTIFICATE is not allowed, users cannot access the CA connectors
        set_policy("user", scope=SCOPE.USER, action=ACTION.AUDIT)
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = json.loads(res.data.decode('utf8')).get("result")
            self.assertIn("enrollCERTIFICATE", result["error"]["message"])

        # with an enrollCERTIFICATE action, users can access the CA connectors
        set_policy("user", scope=SCOPE.USER, action="enrollCERTIFICATE")
        with self.app.test_request_context("/system/names/caconnector",
                                           method="GET",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            _check_caconnector_response(res)

        delete_policy("user")
        delete_policy("admin")
        delete_caconnector("localCA")

    def test_30_realms_with_nodes(self):
        nd1_uuid = "8e4272a9-9037-40df-8aa3-976e4a04b5a9"
        save_resolver({
            "resolver": "local_resolver_1",
            "type": "passwdresolver",
            "file": "/etc/passwd"
        })
        with self.app.test_request_context('/realm/realm_with_node',
                                           method='POST',
                                           json={"resolvers": ["local_resolver_1"],
                                                 "node.local_resolver_1": nd1_uuid,
                                                 "priority.local_resolver_1": 10},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(len(result["value"].get("added")), 1, result)
        delete_realm("realm_with_node")
        delete_resolver("local_resolver_1")

    def test_31_realm_node_api(self):
        nd1_uuid = "8e4272a9-9037-40df-8aa3-976e4a04b5a9"
        nd2_uuid = "d1d7fde6-330f-4c12-88f3-58a1752594bf"
        save_resolver({
            "resolver": "local_resolver_1",
            "type": "passwdresolver",
            "file": "/etc/passwd"
        })
        save_resolver({
            "resolver": "local_resolver_2",
            "type": "passwdresolver",
            "file": "/etc/passwd"
        })
        save_resolver({
            "resolver": "global_resolver",
            "type": "passwdresolver",
            "file": "/etc/passwd"
        })
        # first the request without the node uuid in the database
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json=[{"name": "local_resolver_1",
                                                  "priority": 10}],
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)
            self.assertEqual('ERR905: The given node does not exist!',
                             res.json.get("result").get("error").get("message"),
                             res.json)

        # add the node name and uuid to the database
        db.session.add(NodeName(id=nd1_uuid, name="Node1"))
        db.session.add(NodeName(id=nd2_uuid, name="Node2"))
        db.session.commit()

        # try a weird priority value
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "local_resolver_1",
                                                               "priority": "foo"}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)
            self.assertEqual('ERR905: Could not verify data in request!',
                             res.json.get("result").get("error").get("message"),
                             res.json)
        # missing resolver name
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json={"resolver": [{"priority": "10"}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400, res)
            self.assertEqual('ERR905: Could not verify data in request!',
                             res.json.get("result").get("error").get("message"),
                             res.json)

        # try to add a non-existing resolver
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "unknown_resolver",
                                                               "priority": 10}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(0, len(result["value"].get("added")), result)
            self.assertEqual(1, len(result["value"].get("failed")), result)
            self.assertIn("unknown_resolver", result["value"]["failed"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        self.assertEqual(0, len(realm["realm_with_node"]["resolver"]), realm)

        # add a resolver to the realm
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "local_resolver_1",
                                                               "priority": 10}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(1, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_1", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        reso1 = next(r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_1")
        self.assertEqual(10, reso1["priority"], reso1)
        self.assertEqual(nd1_uuid, reso1["node"], reso1)

        # add the same realm on a different node with no priority
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd2_uuid}',
                                           method='POST',
                                           json={"resolver": [{
                                               "name": "local_resolver_1",
                                               "priority": None}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(2, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_1", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        res_list = [r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_1"]
        # there should be two entries with the same resolver name
        self.assertEqual(2, len(res_list), res_list)
        # both node uuids should be in the resolver list
        reso1n1 = next(r for r in res_list if r["node"] == nd1_uuid)
        reso1n2 = next(r for r in res_list if r["node"] == nd2_uuid)
        # check the corresponding priorities
        self.assertEqual(10, reso1n1["priority"], reso1n1)
        self.assertEqual(None, reso1n2["priority"], reso1n2)

        # update priority of the resolver on the second node
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd2_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "local_resolver_1",
                                                               "priority": "5"}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(2, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_1", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        res_list = [r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_1"]
        # there should be two entries with the same resolver name
        self.assertEqual(2, len(res_list), res_list)
        # both node uuids should be in the resolver list
        reso1n1 = next(r for r in res_list if r["node"] == nd1_uuid)
        reso1n2 = next(r for r in res_list if r["node"] == nd2_uuid)
        # check the corresponding priorities
        self.assertEqual(10, reso1n1["priority"], reso1n1)
        self.assertEqual(5, reso1n2["priority"], reso1n2)

        # add a second resolver on the second node. We need to specify the
        # existing resolver as well otherwise it would be removed
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd2_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "local_resolver_1",
                                                               "priority": "5"},
                                                              {"name": "local_resolver_2",
                                                               "priority": "20"}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(3, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_1", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        res_list = [r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_1"]
        # there should be two entries with the same resolver name
        self.assertEqual(2, len(res_list), res_list)
        # both node uuids should be in the resolver list
        reso1n1 = next(r for r in res_list if r["node"] == nd1_uuid)
        reso1n2 = next(r for r in res_list if r["node"] == nd2_uuid)
        # check the corresponding priorities
        self.assertEqual(10, reso1n1["priority"], reso1n1)
        self.assertEqual(5, reso1n2["priority"], reso1n2)
        reso2n2 = next(r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_2")
        self.assertEqual(20, reso2n2["priority"], reso2n2)

        # remove priority on resolver on node 1
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "local_resolver_1"}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(3, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_1", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        reso1 = next(r for r in realm["realm_with_node"]["resolver"] if r["node"] == nd1_uuid)
        self.assertEqual("local_resolver_1", reso1["name"], reso1)
        self.assertEqual(None, reso1["priority"], reso1)

        # add an unspecific resolver to the realm
        with self.app.test_request_context('/realm/realm_with_node',
                                           method='POST',
                                           json={"resolvers": "global_resolver"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(4, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("global_resolver", result["value"]["added"], result)

        realm = get_realms()
        reso1 = next(r for r in realm["realm_with_node"]["resolver"] if not r["node"])
        self.assertEqual("global_resolver", reso1["name"], reso1)
        self.assertEqual(None, reso1["priority"], reso1)

        # remove resolver_1 on node 2
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd2_uuid}',
                                           method='POST',
                                           json={"resolver": [{"name": "local_resolver_2",
                                                               "priority": "20"}]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(3, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_1", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        res_list = [r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_1"]
        # there should be two entries with the same resolver name
        self.assertEqual(1, len(res_list), res_list)
        # both node uuids should be in the resolver list
        reso1n1 = next(r for r in res_list if r["node"] == nd1_uuid)
        # check the corresponding priorities
        self.assertEqual(None, reso1n1["priority"], reso1n1)
        reso2n2 = next(r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_2")
        self.assertEqual(20, reso2n2["priority"], reso2n2)

        # remove resolver on node 1
        with self.app.test_request_context(f'/realm/realm_with_node/node/{nd1_uuid}',
                                           method='POST',
                                           json={"resolver": []},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result["status"], result)
            self.assertEqual(2, len(result["value"].get("added")), result)
            self.assertEqual(0, len(result["value"].get("failed")), result)
            self.assertIn("local_resolver_2", result["value"]["added"], result)

        realm = get_realms()
        self.assertIn("realm_with_node", realm, realm)
        res_list = [r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_1"]
        # there should be two entries with the same resolver name
        self.assertEqual(0, len(res_list), res_list)
        # both node uuids should be in the resolver list
        reso2n2 = next(r for r in realm["realm_with_node"]["resolver"] if r["name"] == "local_resolver_2")
        self.assertEqual(20, reso2n2["priority"], reso2n2)

        delete_realm("realm_with_node")
        delete_resolver("local_resolver_1")
        delete_resolver("local_resolver_2")



from typing import Union

from webauthn.helpers import bytes_to_base64url

from privacyidea.api.lib.utils import get_required_one_of, get_optional_one_of, get_required
from privacyidea.lib import fido2
from privacyidea.lib.config import get_from_config
from privacyidea.lib.crypto import geturandom
from privacyidea.lib.error import ResourceNotFoundError, AuthError
from privacyidea.lib.fido2.config import FIDO2ConfigOptions
from privacyidea.lib.token import log
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.tokens.passkeytoken import PasskeyTokenClass
from privacyidea.models import Challenge


def get_fido2_nonce() -> str:
    """
    Generate a random 32 byte nonce for fido2 challenges. The nonce is encoded in base64url.
    """
    return bytes_to_base64url(geturandom(32))


def create_fido2_challenge(rp_id: str, user_verification: str = "preferred", transaction_id: Union[str, None] = None,
                           serial: Union[str, None] = None) -> dict:
    """
    Returns a fido2 challenge. If a serial is provided, the challenge is bound to the token with the serial. By default,
    the challenge is not bound to a token, which is the general use-case of FIDO2.
    The challenge validity time is set to either WebauthnChallengeValidityTime, DefaultChallengeValidityTime or 120s
    in that order of evaluation.
    The user_verification parameter can be one of "required", "preferred" or "discouraged". If the value is not one of
    these, "preferred" is used. user_verification is saved in the challenge data field.
    The returned dict has the format:
        ::

            {
                "transaction_id": "12345678901234567890",
                "challenge": <32 random bytes base64url encoded>,
                "rpId": "example.com",
                "message": "Please authenticate with your Passkey!",
                "userVerification": "preferred"
            }
    """
    challenge = fido2.challenge.get_fido2_nonce()
    message = PasskeyTokenClass.get_default_challenge_text_auth()
    validity = int(get_from_config(FIDO2ConfigOptions.CHALLENGE_VALIDITY_TIME,
                                   get_from_config('DefaultChallengeValidityTime', 120)))
    user_verification_values = ["required", "preferred", "discouraged"]
    if user_verification not in user_verification_values:
        log.warning(f"Invalid user_verification value {user_verification}. Using 'preferred' instead.")
        user_verification = "preferred"

    db_challenge = Challenge(serial or "", transaction_id=transaction_id, challenge=challenge,
                             data=f"user_verification={user_verification}", validitytime=validity)
    db_challenge.save()
    transaction_id = db_challenge.transaction_id
    return {
        "transaction_id": transaction_id,
        "challenge": challenge,
        "message": message,
        "rpId": rp_id,
        "user_verification": user_verification
    }


def verify_fido2_challenge(transaction_id: str, token: TokenClass, params: dict) -> int:
    """
    Verify the response for a fido2 challenge with the given token.
    Params is required to have the keys:
    - authenticatorData or authenticatordata
    - clientDataJSON or clientdata
    - signature or signaturedata
    - userHandle or userhandle
    - HTTP_ORIGIN

    If no challenge is found for the transaction_id, a ResourceNotFoundError is raised.
    If the challenge has timed out, an AuthError is raised.
    If the challenge is bound to a token serial and the token serial does not match the input token, an AuthError
    is raised.
    """
    db_challenges = Challenge.query.filter(Challenge.transaction_id == transaction_id).all()
    if not db_challenges:
        raise ResourceNotFoundError(f"Challenge with transaction_id {transaction_id} not found.")

    challenge = next((db_challenge for db_challenge in db_challenges if
                      (db_challenge.serial == token.get_serial() or not db_challenge.serial)),
                     None)
    if not challenge:
        log.error(f"Challenge with transaction_id {transaction_id} is not meant for token {token.get_serial()}.")
        raise AuthError(f"The challenge {transaction_id} is not meant for the token {token.get_serial()}.")

    if not challenge.is_valid():
        log.error(f"Challenge with transaction_id {transaction_id} has timed out.")
        raise AuthError(f"The challenge {transaction_id} has timed out.")

    # Get the user_verification requirement from the challenge data
    uv_string = challenge.get_data()
    parts = uv_string.split("=")
    if len(parts) != 2 or parts[0] != "user_verification":
        log.error(f"Invalid user_verification data in challenge with transaction_id {transaction_id}.")
        raise AuthError(f"Invalid user_verification data in challenge {transaction_id}.")
    user_verification = parts[1]
    if user_verification not in ["required", "preferred", "discouraged"]:
        log.error(
            f"Invalid user_verification value {user_verification} in challenge with transaction_id {transaction_id}."
        )
        raise AuthError(f"Invalid user_verification value {user_verification} in challenge {transaction_id}.")

    options = {
        "challenge": challenge.challenge,
        "authenticatorData": get_required_one_of(params, ["authenticatorData", "authenticatordata"]),
        "clientDataJSON": get_required_one_of(params, ["clientDataJSON", "clientdata"]),
        "signature": get_required_one_of(params, ["signature", "signaturedata"]),
        "userHandle": get_optional_one_of(params, ["userHandle", "userhandle"]),
        "HTTP_ORIGIN": get_required(params, "HTTP_ORIGIN"),
        "user_verification": user_verification
    }
    # These parameters are required for compatibility with the old WebAuthnToken class
    if token.type == "webauthn":
        options.update({"credential_id": get_required_one_of(params, ["credential_id", "credentialid"])})
    options.update({"user": token.user})
    ret = token.check_otp(None, options=options)
    # On success, remove all challenges with the transaction_id
    if ret > 0:
        for db_challenge in db_challenges:
            db_challenge.delete()
    return ret



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import string

from ...lib.error import (ParameterError,
                          AuthError, ERROR)
from ...lib.log import log_with
from privacyidea.lib import _
from privacyidea.lib.utils import (prepare_result, get_version, to_unicode,
                                   get_plugin_info_from_useragent)
import time
import logging
import json
import jwt
import threading
import re
from copy import copy
from urllib.parse import unquote
from flask import (jsonify,
                   current_app, Response)

log = logging.getLogger(__name__)
ENCODING = "utf-8"
TRUSTED_JWT_ALGOS = ["ES256", "ES384", "ES512",
                     "RS256", "RS384", "RS512",
                     "PS256", "PS384", "PS512"]

# The following user-agents (with versions) do not need extra unquoting
# TODO: we should probably switch this when we do not do the extra unquote anymore
NO_UNQUOTE_USER_AGENTS = {
    'privacyidea-cp': None,
    'privacyidea-ldap-proxy': None,
    'simplesamlphp': None
}

SESSION_KEY_LENGTH = 32

optional = True
required = False


def _get_param(dictionary, key, default=None):
    """
    Get the parameter from the dictionary. If the parameter is not present, return the default value or None.
    """
    ret = None
    if dictionary and key in dictionary:
        ret = dictionary[key]
    elif default is not None:
        ret = default
    return ret


def get_required(dictionary, key):
    """
    Get the required parameter from the dictionary. If the parameter is not present, raise a ParameterError.
    If the parameter is present, but empty, the empty value will be returned.
    This just checks for None.
    """
    ret = _get_param(dictionary, key, None)
    if not ret:
        raise ParameterError(f"Missing parameter: {key}", id=905)
    return ret


def get_required_one_of(param, keys):
    """
    Get the first parameter from the list of keys that is present in the param dictionary.
    If none of the keys is present, raise a ParameterError.
    """
    for key in keys:
        ret = _get_param(param, key, None)
        if ret:
            return ret
    raise ParameterError(f"Missing one of the following parameters: {keys}", id=905)


def get_optional(param, key, default=None):
    """
    Get the optional parameter from the dictionary. If the parameter is not present, return the default value or None.
    """
    return _get_param(param, key, default)


def get_optional_one_of(param, keys, default=None):
    """
    Get the first parameter from the list of keys that is present in the param dictionary.
    If none of the keys is present, return the default value or None.
    """
    for key in keys:
        ret = _get_param(param, key, default)
        if ret is not None:
            return ret
    return default


def getParam(param, key, optional=True, default=None, allow_empty=True, allowed_values=None):
    """
    returns a parameter from the request parameters.

    :param param: the dictionary of parameters
    :type param: dict
    :param key: the name of the parameter
    :param optional: defines if this parameter is optional or not
                     an exception is thrown if the parameter is required
                     otherwise: nothing done!
    :type optional: bool
    :param default: The value to assign to the parameter, if it is not
                    contained in the param.
    :param allow_empty: Set to False is the parameter is a string and is
        not allowed to be empty
    :param allowed_values: A list of allowed values. If another value is given,
        then the default value is returned
    :type allow_empty: bool

    :return: the value (literal) of the parameter if exists or nothing
             in case the parameter is optional, otherwise throw an exception
    """
    ret = None

    if key in param:
        ret = param[key]
    elif default:
        ret = default
    elif not optional:
        raise ParameterError("Missing parameter: {0!r}".format(key), id=905)

    if not allow_empty and ret == "":
        raise ParameterError("Parameter {0!r} must not be empty".format(key), id=905)

    if allowed_values and ret not in allowed_values:
        ret = default

    return ret


def send_result(obj, rid=1, details=None) -> Response:
    """
    sendResult - return a json result document

    :param obj: simple result object like dict, sting or list
    :type obj: dict or list or string/unicode
    :param rid: id value, for future versions
    :type rid: int
    :param details: optional parameter, which allows to provide more detail
    :type  details: None or simple type like dict, list or string/unicode

    :return: json rendered string result
    :rtype: string
    """
    return jsonify(prepare_result(obj, rid, details))


def send_error(errstring, rid=1, context=None, error_code=-311, details=None):
    """
    sendError - return a json error result document

    remark:
     the 'context' is especially required to catch errors from the _before_
     methods. The return of a _before_ must be of type response and
     must have the attribute response._exception set, to stop further
     processing, which otherwise will have ugly results!!

    :param errstring: An error message
    :type errstring: basestring
    :param rid: id value, for future versions
    :type rid: int
    :param context: default is None or 'before'
    :type context: string
    :param error_code: The error code in the JSON object along with the error
    message.
    :type error_code: int
    :param details: dict with additional details about the error (like
        challenges)
    :type details: dict

    :return: json rendered sting result
    :rtype: string

    """
    if details:
        details["threadid"] = threading.current_thread().ident
    res = {"jsonrpc": "2.0",
           "detail": details,
           "result": {"status": False,
                      "error": {"code": error_code,
                                "message": errstring}
                      },
           "version": get_version(),
           "id": rid,
           "time": time.time()
           }

    ret = jsonify(res)
    return ret


def send_html(output):
    """
    Send the output as HTML to the client with the correct mimetype.

    :param output: The HTML to send to the client
    :type output: str
    :return: The generated response
    :rtype: flask.Response
    """
    return current_app.response_class(output, mimetype='text/html')


def send_file(output, filename, content_type='text/csv'):
    """
    Send the output to the client with the "Content-disposition" header to
    declare it as a downloadable file.

    :param output: The data that should be sent as a file
    :type output: str
    :param filename: The proposed filename
    :type filename: str
    :param content_type: The proposed content type of the data
    :type content_type: str (should be something from this list:
                             https://www.iana.org/assignments/media-types/media-types.xhtml)
    :return: The generated response
    :rtype: flask.Response
    """
    headers = {'Content-disposition': 'attachment; filename={0!s}'.format(filename)}
    return current_app.response_class(output, headers=headers, mimetype=content_type)


def send_csv_result(obj, data_key="tokens",
                    filename="privacyidea-tokendata.csv"):
    """
    returns a CSV document of the input data (like in /token/list)

    It takes an obj as a dict like:
    { "tokens": [ { ...token1... }, { ...token2....}, ... ],
      "count": 100,
      "....": .... }

    :param obj: The data, that gets serialized as CSV
    :type obj: dict
    :param data_key: The key, from which the list should be returned as CSV.
    Usually this is "tokens".
    :type data_key: basestring
    :param filename: The filename to save the CSV to.
    :type filename: basestring
    :return: The result serialized as a CSV
    :rtype: Response object
    """
    delim = "'"
    output = ""
    # check if there is any data
    if data_key in obj and len(obj[data_key]) > 0:
        # Do the header
        for k, _v in obj.get(data_key)[0].items():
            output += "{0!s}{1!s}{2!s}, ".format(delim, k, delim)
        output += "\n"

        # Do the data
        for row in obj.get(data_key):
            for val in row.values():
                if isinstance(val, str):
                    value = val.replace("\n", " ")
                else:
                    value = val
                output += "{0!s}{1!s}{2!s}, ".format(delim, value, delim)
            output += "\n"

    return send_file(output, filename)


@log_with(log)
def getLowerParams(param):
    ret = {}
    for key in param:
        lkey = key.lower()
        # strip the session parameter!
        if "session" != lkey:
            lval = param[key]
            ret[lkey] = lval
    return ret


def check_unquote(request, data):
    """
    Check if we need to unquote the given data.
    Based on the user-agent header of the request we unquote the given values
    in `data`. The user-agent string parsing is based on
    https://httpwg.org/specs/rfc9110.html#field.user-agent

    :param request: The Flask request context
    :type request: Flask.Request
    :param data: The dictionary containing the requested data
    :type data: dict
    :return: New dictionary with the possibly unquoted values
    :rtype: dict
    """
    ua_name, ua_version, _ua_comment = get_plugin_info_from_useragent(request.user_agent.string)
    # if no user agent is available, we assume that we must unquote the data
    if not ua_name:
        return {key: unquote(value) for (key, value) in data.items()}

    if ua_name.lower() not in NO_UNQUOTE_USER_AGENTS:
        return {key: unquote(value) for (key, value) in data.items()}
    else:
        return copy(data)


def get_all_params(request):
    """
    Retrieve all parameters from a request, no matter if these are GET or POST requests
    or parameters are contained as viewargs like the serial in DELETE /token/<serial>

    :param request: The flask request object
    """
    param = request.values
    body = request.data
    return_param = {}
    if param:
        log.debug("Update params in request {0!s} {1!s} with values.".format(request.method,
                                                                             request.base_url))
        # Add the unquoted HTML and form parameters
        return_param = check_unquote(request, request.values)

    if request.is_json:
        log.debug("Update params in request {0!s} {1!s} with JSON data.".format(request.method,
                                                                                request.base_url))
        # Add the original JSON data
        return_param.update(request.json)
    elif body:
        # In case of serialized JSON data in the body, add these to the values.
        try:
            json_data = json.loads(to_unicode(body))
            for k, v in json_data.items():
                return_param[k] = v
        except Exception as exx:
            log.debug("Can not get param: {0!s}".format(exx))

    if request.view_args:
        log.debug("Update params in request {0!s} {1!s} with view_args.".format(request.method,
                                                                                request.base_url))
        # We add the unquoted view_args
        return_param.update(check_unquote(request, request.view_args))

    return return_param


def get_priority_from_param(param):
    """
    Return a dictionary of priorities as int from params like
    priority.key1=value1

    :param param: The params dictionary
    :type param: dict
    :return: dict
    """
    priority = {}
    for k, v in param.items():
        if k.startswith("priority.") and isinstance(v, int):
            priority[k[len("priority."):]] = int(v)
    return priority


def verify_auth_token(auth_token, required_role=None):
    """
    Check if a given auth token is valid.

    Return a dictionary describing the authenticated user.

    :param auth_token: The Auth Token
    :param required_role: list of "user" and "admin"
    :return: dict with authtype, realm, rights, role, username, exp, nonce
    :rtype: dict
    """
    r = None
    if required_role is None:
        required_role = ["admin", "user"]
    if auth_token is None:
        raise AuthError(_("Authentication failure. Missing Authorization header."), id=ERROR.AUTHENTICATE_AUTH_HEADER)

    try:
        headers = jwt.get_unverified_header(auth_token)
    except jwt.DecodeError as err:
        raise AuthError(_("Authentication failure. Error decoding the Authorization token:") + f" {err!s}",
                        id=ERROR.AUTHENTICATE_DECODING_ERROR)
    algorithm = headers.get("alg")
    wrong_username = None
    if algorithm in TRUSTED_JWT_ALGOS:
        # The trusted JWTs are RSA, PSS or elliptic curve signed
        trusted_jwts = current_app.config.get("PI_TRUSTED_JWT", [])
        for trusted_jwt in trusted_jwts:
            try:
                if trusted_jwt.get("algorithm") in TRUSTED_JWT_ALGOS:
                    j = jwt.decode(auth_token, trusted_jwt.get("public_key"), algorithms=[trusted_jwt.get("algorithm")])
                    if (dict((k, j.get(k)) for k in ("role", "resolver", "realm")) ==
                            dict((k, trusted_jwt.get(k)) for k in ("role", "resolver", "realm"))):
                        if re.match(trusted_jwt.get("username") + "$", j.get("username")):
                            r = j
                            break
                        else:
                            r = wrong_username = j.get("username")
                else:
                    log.warning("Unsupported JWT algorithm in PI_TRUSTED_JWT.")
            except jwt.DecodeError as _e:
                log.info("A given JWT definition does not match.")
            except jwt.ExpiredSignatureError as err:
                # We have the correct token. It expired, so we raise an error
                raise AuthError(_("Authentication failure. Your token has expired:") + f" {err!s}",
                                id=ERROR.AUTHENTICATE_TOKEN_EXPIRED)

    if not r:
        try:
            r = jwt.decode(auth_token, current_app.secret_key, algorithms=['HS256'])
        except jwt.DecodeError as err:
            raise AuthError(_("Authentication failure. Error decoding the Authorization token:") + f" {err!s}",
                            id=ERROR.AUTHENTICATE_DECODING_ERROR)
        except jwt.ExpiredSignatureError as err:
            raise AuthError(_("Authentication failure. Your token has expired:") + f" {err!s}",
                            id=ERROR.AUTHENTICATE_TOKEN_EXPIRED)
    if wrong_username:
        raise AuthError(_("Authentication failure. The username {wrong_username} "
                          "is not allowed to impersonate via JWT.").format(wrong_username=wrong_username))
    if required_role and r.get("role") not in required_role:
        # If we require a certain role like "admin", but the users role does
        # not match
        raise AuthError(_("Authentication failure. You do not have the necessary "
                          "role ({required_role}) to access this resource!").format(required_role=required_role),
                        id=ERROR.AUTHENTICATE_MISSING_RIGHT)
    return r


def check_policy_name(name):
    """
    This function checks, if the given name is a valid policy name.

    :param name: The name of the policy
    :return: Raises a ParameterError in case of an invalid name
    """
    disallowed_patterns = [("^check$", re.IGNORECASE),
                           ("^pi-update-policy-", re.IGNORECASE)]
    for disallowed_pattern in disallowed_patterns:
        if re.search(disallowed_pattern[0], name, flags=disallowed_pattern[1]):
            raise ParameterError(_("Invalid policy name:") + f" {name}")

    if not re.match(r'^[a-zA-Z0-9_.\- ]*$', name):
        raise ParameterError(_("The name of the policy may only contain the characters a-zA-Z0-9_. -"))


def attestation_certificate_allowed(cert_info, allowed_certs_pols):
    """
    Check a certificate against a set of policies.

    This will check an attestation certificate of a U2F-, or WebAuthn-Token,
    against a list of policies. It is used to verify, whether a token with the
    given attestation may be enrolled, or authorized, respectively.

    The certificate info may be None, in which case, true will be returned if
    the policies are also empty.

    :param cert_info: The `attestation_issuer`, `attestation_serial`, and `attestation_subject` of the cert.
    :type cert_info: dict or None
    :param allowed_certs_pols: The policies restricting enrollment, or authorization.
    :type allowed_certs_pols: dict or None
    :return: Whether the token should be allowed to complete enrollment, or authorization, based on its attestation.
    :rtype: bool
    """

    if not cert_info:
        return not allowed_certs_pols

    if allowed_certs_pols:
        for allowed_cert in allowed_certs_pols:
            tag, matching, _rest = allowed_cert.split("/", 3)
            tag_value = cert_info.get(f"attestation_{tag!s}")
            # if we do not get a match, we bail out
            m = re.search(matching, tag_value) if matching and tag_value else None
            if matching and not m:
                return False

    return True


def is_fqdn(x):
    """
    Check whether a given string could plausibly be a FQDN.

    This checks, whether a string could be a FQDN. Please note, that this
    function will currently return true for plenty of strings, that are not
    actually valid FQDNs. This is expected. This function performs a simple
    plausibility check to ward against obvious mistakes, like a user
    accidentally putting in a full url with protocol. The caller should not
    rely on this function, if it is absolutely crucial, that the checked
    string is a valid FQDN. It is solely intended to be used to implement user
    convenience, by alerting the user early on, if they have misunderstood
    a particular fields purpose.

    :param x: String to check.
    :type x: basestring
    :return: Whether the given string may plausibly be a FQDN.
    :rtype: bool
    """
    return set(string.punctuation).intersection(x).issubset({'-', '.'})
